Package: Rook
Title: HTTP Web Server for R
Version: 1.2
Date: 2022-11-05
Description: An HTTP web server for R with a documented API to interface between R and the server. The documentation contains the Rook specification and details for building and running Rook applications. To get started, be sure and read the 'Rook' help file first.
Encoding: UTF-8
Depends: R (>= 2.13.0)
Imports: utils, tools, methods, brew
License: GPL-2
LazyLoad: yes
URL: https://github.com/evanbiederstedt/rook
BugReports: https://github.com/evanbiederstedt/rook/issues
Author: Jeffrey Horner [aut], Evan Biederstedt [aut, cre]
Maintainer: Evan Biederstedt <evan.biederstedt@gmail.com>
NeedsCompilation: yes
Packaged: 2022-11-05 23:55:46 UTC; evanbiederstedt
Repository: CRAN
Date/Publication: 2022-11-07 08:50:19 UTC
Built: R 4.4.0; amd64-portbld-midnightbsd4.0.2; 2026-02-18 02:42:31 UTC; unix

Package: Rook
Title: HTTP Web Server for R
Version: 1.2
Date: 2022-11-05
Description: An HTTP web server for R with a documented API to interface between R and the server. The documentation contains the Rook specification and details for building and running Rook applications. To get started, be sure and read the 'Rook' help file first.
Encoding: UTF-8
Depends: R (>= 2.13.0)
Imports: utils, tools, methods, brew
License: GPL-2
LazyLoad: yes
URL: https://github.com/evanbiederstedt/rook
BugReports: https://github.com/evanbiederstedt/rook/issues
Author: Jeffrey Horner [aut], Evan Biederstedt [aut, cre]
Maintainer: Evan Biederstedt <evan.biederstedt@gmail.com>
NeedsCompilation: yes
Packaged: 2022-11-05 23:55:46 UTC; evanbiederstedt
Repository: CRAN
Date/Publication: 2022-11-07 08:50:19 UTC
Built: R 4.4.0; i386-portbld-midnightbsd4.0; 2026-05-15 06:07:18 UTC; unix

Package: Rook
Version: 1.2.1
Title: HTTP Web Server for R
Authors@R: c(person(given = "Jeffrey",
                    family = "Horner",
                    role = "aut"),
             person(given = "Evan",
                    family = "Biederstedt",
                    role = c("aut", "cre"),
                    email = "evan.biederstedt@gmail.com"))
Date: 2026-06-30
Description: An HTTP web server for R with a documented API to interface between R and the server. The documentation contains the Rook specification and details for building and running Rook applications. To get started, be sure and read the 'Rook' help file first.
Encoding: UTF-8
Depends: R (>= 2.13.0)
Imports: utils, tools, methods, brew
License: GPL-2
LazyLoad: yes
URL: https://github.com/evanbiederstedt/rook
BugReports: https://github.com/evanbiederstedt/rook/issues
Maintainer: Evan Biederstedt <evan.biederstedt@gmail.com>
NeedsCompilation: yes
Packaged: 2026-06-30 19:30:21 UTC; evanbiederstedt
Author: Jeffrey Horner [aut],
  Evan Biederstedt [aut, cre]
Repository: CRAN
Date/Publication: 2026-07-01 00:00:02 UTC

exportClass(Rhttpd,RhttpdApp,RhttpdInputStream,RhttpdErrorStream)
export(Rhttpd,RhttpdApp,RhttpdInputStream,RhttpdErrorStream)
exportClass(Utils,Multipart)
export(Utils,Multipart)
exportClass(Request,Response,Mime)
export(Request,Response,Mime)
exportClass(File,Static,Builder,App,Middleware,URLMap)
export(File,Static,Builder,App,Middleware,URLMap)
exportClass(Brewery,Redirect)
export(Brewery,Redirect)
export(is_rookable)
export(Server,suspend_console)
useDynLib(Rook,.registration=TRUE)
import(tools)
import(utils)
import(methods)
import(brew)

[![<evanbiederstedt>](https://circleci.com/gh/evanbiederstedt/Rook.svg?style=svg)](https://app.circleci.com/pipelines/github/evanbiederstedt/Rook)
[![CRAN status](https://www.r-pkg.org/badges/version/Rook)](https://cran.r-project.org/package=Rook)
[![CRAN downloads](https://cranlogs.r-pkg.org/badges/Rook)](https://cran.r-project.org/package=Rook)


**NOTE:** This is a fork of the original project at https://github.com/jeffreyhorner/Rook

The original maintainer didn't respond to fix various issues which required the package to be removed from CRAN. 

This fork fixes those issues. As a consequence, I ([@evanbiederstedt](https://github.com/evanbiederstedt)) am the new maintainer. 


Rook: A web server interface for R
=======================================

This specification defines the interface between web servers and R applications.

Rook applications
-----------------

A Rook application is an R reference class object that implements a 'call'
method or an R closure that takes exactly one argument, an environment,
and returns a list with three named elements: the 'status', the 'headers',
and the 'body'.

Hello World
-----------

Here is a basic Rook application as a closure that implements 'hello world':

    function(env){
	body = paste('<h1>Hello World! This is Rook',env$rook.version,'.</h1>')
	list(
	    status = 200L,
	    headers = list(
		'Content-Type' = 'text/html'
	    ),
	    body = body
	)
    }

And the equivalent reference class example:

    setRefClass(
	'HelloWorld',
	methods = list(
	    call = function(env){
		body = paste('<h1>Hello World! This is Rook',env$rook.version,'.</h1>')
		list(
		    status = 200L,
		    headers = list(
			'Content-Type' = 'text/html'
		    ),
		    body = body
		)
	    }
	)
    )

The Environment
---------------

The environment argument is a true R environment object which the
application is free to modify. It is required to contain the following
variables:

- REQUEST_METHOD: The HTTP request method, such as "GET" or "POST". This
    cannot ever be an empty string, and so is always required.

- SCRIPT_NAME:	The initial portion of the request URL‘s "path" that
    corresponds to the application object, so that the application knows
    its virtual "location". This may be an empty string, if the application
    corresponds to the "root" of the server.

- PATH_INFO:  The remainder of the request URL‘s "path", designating the
    virtual "location" of the request‘s target within the application. This
    may be an empty string, if the request URL targets the application root
    and does not have a trailing slash. This value may be percent-encoded
    when I originating from a URL.

- QUERY_STRING:	The portion of the request URL that follows the ?,
    if any. May be empty, but is always required!

- SERVER_NAME, SERVER_PORT:   When combined with SCRIPT_NAME and PATH_INFO,
    these variables can be used to complete the URL. Note, however, that
    HTTP_HOST, if present, should be used in preference to SERVER_NAME for
    reconstructing the request URL. SERVER_NAME and SERVER_PORT can never
    be empty strings, and so are always required.

- HTTP_ Variables:    Variables corresponding to the client-supplied
    HTTP request headers (i.e., variables whose names begin with HTTP_). The
    presence or absence of these variables should correspond with the presence
    or absence of the appropriate HTTP header in the request.

In addtion, the environment must include the following Rook-specific variables:

- rook.version:	    This version of Rook.
- rook.url_scheme:    http or https, depending on the request URL.
- rook.input:	    See below, the input stream.
- rook.errors:	    See below, the error stream.

The Input Stream
----------------

The rook.input variable must contain an object created from a reference
class and respond to read_lines, read, and rewind:

read_lines: takes one argument, the number of lines to read. Includes partial ending line.
read: takes one argument, the number of bytes to read. Returns a raw vector.
rewind: Rewinds the input stream back to the beginning.

The Error Stream
----------------

The rook.error variable must contain an object created from a reference
class and must respond to flush and cat:

flush: called with no arguments and makes the error stream immediately appear.
cat: called with the same arguments as R's cat without the file and append argument.

The Response
============

The Status
----------

This is an HTTP status value and must be greater than or equal to 100.

The Headers
-----------

This is a named list that contains string values only corresponding to valid HTTP headers.

The Body
--------

This is either a character or raw vector. If the character vector is
named with value 'file' then value of the vector is interpreted as the
location of a file.

# Rook

![logo](https://hg.sr.ht/~ser/rook/raw/assets/rook.png?rev=tip)

[![Build status](https://builds.sr.ht/~ser/rook/.build.yml.svg)](https://builds.sr.ht/~ser/rook/.build.yml?)
★ [Releases](https://downloads.ser1.net/software/rook/)
★ ![Latest release](https://downloads.ser1.net/software/rook/badges/latest_release.svg)
★ [![AUR version](https://img.shields.io/aur/version/rook)](https://aur.archlinux.org/packages/rook)  
[![License](https://img.shields.io/aur/license/rook)](https://hg.sr.ht/~ser/rook/raw/LICENSE?rev=tip)
★ [Changelog](https://hg.sr.ht/~ser/rook/browse/CHANGELOG.md?rev=tip)
★ [Software metrics](https://downloads.ser1.net/software/rook/badges/METRICS.md)
★ [Issues](https://todo.sr.ht/~ser/rook)
★ [Mailing List](https://lists.sr.ht/~ser/rook)  
[![Go Report Card](https://goreportcard.com/badge/ser1.net/rook)](https://goreportcard.com/report/ser1.net/rook)
★ ![Code linting](https://downloads.ser1.net/software/rook/badges/linting.svg)
★ ![Code security](https://downloads.ser1.net/software/rook/badges/vulnerabilities.svg)
★ ![Markdown linting](https://downloads.ser1.net/software/rook/badges/readme_linting.svg)
★ ![Unit tests](https://downloads.ser1.net/software/rook/badges/unit_tests.svg)  
![Code coverage](https://downloads.ser1.net/software/rook/badges/coverage.svg)
★ ![Code complexity](https://downloads.ser1.net/software/rook/badges/complexity.svg)
★ ![ABC metrics](https://downloads.ser1.net/software/rook/badges/abc_score.svg)
★ ![Lines of code](https://downloads.ser1.net/software/rook/badges/loc.svg)

A lightweight, stand-alone, headless secret service tool backed by a Keepass v2
database.

- [Usage](#usage)
- [Install](#install)
- [Security](SECURITY.md)
- [Roadmap](#roadmap)
- [Contributing](CONTRIBUTING.md)
- [Credits and Related](#credits-and-related)
- [Other solutions](#other-solutions)

Rook allows you to use a KeePass v2 database as storage for secrets. It provides
client and server modes; the server unlocks the database and stays in memory;
the client communicates over a socket with the server and fetches data. This
allows:

- Using your KeePass database, instead of having to have all your secrets _also_
  stored in a second system
- Allows you to unlock the database once, instead of every time the system needs
  a secret
- Allows you to use your secrets in scripts and tools such as
  [mbsync/isync](https://github.com/gburd/isync),
  [vdirsyncer](https://github.com/pimutils/vdirsyncer),
  [offlineimap](https://github.com/OfflineIMAP/offlineimap),
  [msmtp](https://marlam.de/msmtp), encrypted file system mounters, encrypted
  and/or offsite backup programs, IMAP email clients, and so on.
- Allows you to run a _headless_ secret server
- Is easily auditable, and as a single binary is more easily protectable

There are other ways of accomplishing each of these, but none that accomplish
all of them. passhole and kpmenu are close alternatives. passhole is Python with
a lot of dependencies to audit. Recent versions of kpmenu also have large
dependency trees, and the mode of operation makes it harder to use for non-GUI
scripting.

Rook is not compatible with any other secret service tools; it does not
communicate over, or rely on, DBus, or anything systemd. It should work as well
on Darwin, *BSD, or any other POSIX-ish OS as it does on Linux, although this
has not been tested. I don't know what it'll do on Windows.

## Status & Features

Rook is mostly complete. Some features are missing; see the Roadmap section
below. It might crash, although it's been a while since I've seen one. It should
be currently able to replace most secret-service type calls you might make from
scripts; you can easily fetch passwords by account with the client, which is
what you really need.

- Read kdbx v2 Keeass database files
- List entries from the DB
- Search entries, by title, URL, or tag
- Show entries
- Show individual fields, such as username & password
- Reload a (presumably changed) database; open a different database
- Lock / unlock a database
- Pin-controlled client calls for additional security (see Secure section,
  below)
- Headless
- One-time pin soft locking & unlocking
- Minimal dependencies and an auditable amount of sourcecode -- you can check to
  make sure it's not doing anything fishy, even if you aren't a Go expert. This
  repo contains around 1k LOC in 5 files. Assuming you trust the core Go team,
  you only then have to audit two other external libraries: a [flags
  lib](https://hg.sr.ht/~ser/claptrap) written by me that is a single Go file
  with no external dependencies and which has no networking code; and [the
  library](https://github.com/tobischo/gokeepasslib) that provides access to the
  KeepassDB (gokeepasslib has itself one external dependency, a crypto library
  written by the same author) - neither of which have networking code.
- Supports either password, or password + keyfile credentials

## Usage

`rook help` prints a list of the supported commands. Start a rook server process
with `rook serve`, and then access it by calling `rook` again with one of the
other commands. There is no configuration, or configuration file.

Exit the rook server with `rook exit` or `^c`. If, for some reason, it leaves
behind the socket file, the next time you start rook it'll complain about one
already running. You can override that warning with the `-f` argument.

Rook will watch the DB file and auto-refresh the memory cache if the file
changes.

### Serving a database

Start a rook server with:

```bash
rook -f serve -P /path/to/your.kdbx
```

`-f` here forces rook to ignore and overwrite any pre-existing socket file; you
probably shouldn't use it unless you are certain rook isn't already running.
`-P` tells rook to prompt for the database password. `-P` can be used either
when starting a server, or when calling `open` from a client. Alternatively
(and, probably, more normally), start the rook server with no password, and
unlock it with a client call when you can prompt for a password:

```bash
$ rook serve /path/to/kdbx &
2025/04/08 10:40:57 [ERROR] unable to decode database "/path/to/kdbx": Wrong password? HMAC-SHA256 of header mismatching
2025/04/08 10:40:57 [ERROR] reload() error on server creation
$ rook open -P
Password:
```

The errors in the `serve` command is because rook is unable to open the database
as no password has been provided. You can avoid this error by running it
interactively like this:

```
$ rook serve -P /path/to/kdbx
Password:
Rook server started with "/path/to/kdbx"
```

This obviously can't be done if you're auto-starting rook as part of your login
process. If you both want to autostart rook _and_ start it unlocked, you can use
a command like [zenity](https://gitlab.gnome.org/GNOME/zenity) to prompt for the
password:

```
zenity --password --title "Unlock rook" | rook serve --stdin /path/to/kdbx
```

If you use a key file for a second security factor, you can specify it with the
`-k` argument to either `serve` or `open`.

Or, the server can be called with no database, and the client can tell it which
database to open by passing it a path; in this case, it must be a path the
server can find the file through, usually a fully qualified path. However, if
the rook server was run in the same directory as one or more databases, the
client can provide a simple relative path:

```bash
$ rook open -P my.kdbx
Password:
SUCCESS
```

A database supplied by `open` replaces whatever the server was started with.

As an alternative to all of password prompting, rook will also take a password
from the `ROOK_PASSWORD` environment variable. The prompt takes precedence.
Remember that passwords are plain text, and environment variables are -- while
reasonably secure -- still more persistent and accessible than other mechanisms.
The password may also be provided to the `serve` command via stdin pipe with the
`--stdin` argument; this allows starting a Rook server with a password vault
without exposing the password through environment variables and without
requiring user interaction. This means you could also do this in an interactive
bash, as an alternative to `-P`:

```
$ rook serve --stdin database.kdbx
supersecretpassword
^d
```

It's possible to tell whether the database is unlocked by making most any other
request, e.g. `info`, as the server will return a `NO OPEN DATABASE` error
message. If no database is open, the `lock` and `commands` requests will not
throw errors, and the only other request that will work is `open`.

Databases can be live-reloaded with the `reload` command; this assumes the
credentials have not changed,and it simply replaces the in-memory database from
the current version on disk. There is not yet -- and may never be -- fancy DB
merging such as KeepassXC does; it's a straight-up replacement.

If you want to move the socket location, e.g. to `/var/run/user/$(id
-u)/rook.sock` then use the `--socket` argument. If you do this, you **must**
set it for **both** the server and every client call. Easiest would be to create
an alias in your shell's rc file, but you can also just specify it everywhere
you call Rook.

### Searching & showing

If the entry Title is unique, then `show` only needs that; otherwise, it'll show
all the matching entities -- which may not be what you want when using rook for
script authentication purposes. In the case of duplicates, provide the path -
separated by `/` - to the entity -- or part of it. For example, if you Keepass
database looks like:

```
Root
  Internet
    Accounts
      me
  Server
    Accounts
      me
  Identities
    me
    myself
```

then `rook show me` will show of all three `me` entries; `rook show Accounts me`
will show two of them; `rook show myself` will show only the one entry; and
`rook show Server/Accounts me` or `rook show Identities me` will show only the
one entry. A fully qualified path would be `rook show Root/Internet/Accounts
me`, but rook requires only enough to uniquely identify entries.

Rook can also display your database as a tree by using the `--tree` argument:

```
$ rook ls -t
┌─ /
├─┬─ Subgroup  1
│ ├─┬─ Subsubgroup 2
│ │ └── Subsub item
│ └── Subgroup 1 item 1
├── First entry
├── Second test
├── Third entry
├── fourth entry
└── Something wicked
```

Empty groups are not shown; leaves in the tree are _always_ entries.

### Locking

Databases can be hard-locked with the `lock` command, e.g. `rook lock`. With a
hard lock, rook forgets everything it knows about the database, except for the
path on the filesystem. To unlock the database, a client call of `rook open -P`
is required —- this will prompt for the DB password.

Databases can also be soft-locked by calling `lock --pin`; in this case, the
rook server will generate a 4-digit one-time pin and return it to the client.
This pin can be used one time to unlock the database with `open --pin <digits>`.
If `open --pin` is called incorrectly 3 times, the database is hard-locked.  Any
other calls to rook while the database is in this state are soft-failures, in
that they return an error, but do not count as failed unlock attempts. This
prevents cron-jobbed scripts looking up paswords from hard-locking the database,
and since none of the requests can be used to brute-force rook, they're
considered mostly harmless. Similarly, attempts to call `lock --pin` while the
database is locked also soft-fail.

If a database is soft-locked, a client call to `open -P` (providing the full
password) will also unlock the database.

The keyring pin may **not** be used as a pin to unlock a locked database. Only a
one-time pin created with `lock --pin` or a full password provided by `-P` may
be used with `open` to unlock a database. Running `--keyring open` will result
in an error (but no server call).

Rook does not have any built-in support for timing-out and locking databases.
This can be accomplished with cron, or systemd timer units. Since the intended
purpose of rook is to provide secrets to cron jobs, such as email and calendar
syncing, there is no "timeout after disuse period" functionality. However, in
some cases a more ideal behavior can be tied into the screen locker; an example
is provided in [Examples](#examples), below.

#### Failures

Some conditions will cause a hard-lock of the database, requiring the full
password to unlock. In all failure cases that count towards a hard lock, the
caller has 3 tries, after which the server will lock and the DB and clear it
from memory. Successful attempts to read the DB clear the ticker.

Attempts to _read_ the database without proper authorization, such as calling
`ls` on a `--secure`d or `--keyring` secured DB count towards a hard-lock.

Failed attempts to open the database, as with the wrong password or pin, count
towards a hard-lock. This hinders brute-forcing the password.

If a database is soft-locked with a one-time pin, reads are _not_ counted as
failures towards a hard-lock; this prevents automated background jobs attempting
to read the DB from triggering a hard-lock. The reads will still fail, but they
won't trigger a hard-lock.

Open attempt failures on a soft-locked DB **do** count towards a hard-lock.

### Improving security

rook knows how to use the kernel keyring to store a pin in the user's session.
Using this feature greatly increases security, as only processes in the user's
session (children of the X or Wayland session) can access this keyring. This
protects other processes -- including root processes -- from sending commands
directly to the socket and accessing secrets.

To use this feature, run the rook server and client with the `--keyring`
argument. Rook will automatically set and fetch the pin from the keyring's user
session.

This can be done manually, using the `keyctl` command from the `keyutils`
package; this process is described in the [Security](SECURITY.md) document.

While `--keyring` automates using the key ring, it uses the kernel API. In
particular, it does not fork out to the `keyctl` command. Using this feature
from within a container is undefined and unsupported -- if you have success with
this, great; if not, please don't bother filing a bug. Container permissions and
settings are beyond the scope of this project.

### Caveats

`rook list` shows all entries in the DB, with an index number and a path. While
the index can be used, it's not a fixed thing and could change as the database
changes, so it's best to not rely on it in scripts.

`rook exit` immediately performms a hard `os.Exit()` without requiring a
password, and which will terminate any other concurrent pending client requests.
This is intended behavior. As with all behavior, rook errs on the side security.
I could be convinced to require an authenticated request; send an email to the
mailing list if you have a compelling argument (or need) for that.

### Examples

The alias in the following example isn't necessary; it's just to keep rook
running everything in `${PWD}`.

![](https://cloud.ser1.net/api/public/dl/a7v3uu5d?inline=true)

Here are some code snippets from various config files (if you're on Linux, I
highly recommend you use the `--keyring` feature; just add it after `rook` in
all the examples):

**mbsync** (`~/.config/isyncrc`)

```
PassCmd "rook show -p 'Email account'"
```

**aerc** (`~/.config/aerc/accounts.conf`)

```
source-cred-cmd = rook show -p Accounts 'Email account' 
```

**vdirsyncer** (`~/.config/vdirsyncer/config`)

```
password.fetch = ["command", "rook","show","-p","CardDAV account"] 
```

**msmtp** (`~/.msmtprc`)

```
passwordeval   rook show -p Identities "SMTP server"
```

**restic** (environment variables)

You could store all of your restic backup information in Keepass:

```
RESTIC_PASSWORD=$(rook show -p Accounts/Backups glamdring) \
  RESTIC_REPOSITORY=$(rook show -f restic_repository Accounts/Backups glamdring) \
  EXCLUDES=($(rook show -f excludes Accounts/Backups glamdring)) \
  /usr/bin/restic backup ${EXCLUDES[@]/#/-e } --one-file-system \
  $(rook show -f files Accounts/Backups glamdring)
```

#### Searching

Searches look in several places for matches: titles, URLs, and tags. However,
`show` only matches against titles. Both `search` and `show` also take an
optional path as the first argument, which filters the entities before matching;
e.g.:

```bash
$ rook ls
1    Accounts   Entry 1
2    Accounts   Entry 2
3    Accounts/Hosting   Entry 1
4    Accounts/Hosting   Entry 2
5    Internet   Entry 1
6    Internet   Entry 2
7    Internet/Hosting   Entry 1
$ rook search 'Entry 2'
Index: 2
Path: Accounts
Title: Entry 2

Index: 4
Path: Accounts/Hosting
Title: Entry 2

Index: 6
Path: Internet
Title: Entry 2
$ rook search Internet 'Entry 2'
Index: 6
Path: Internet
Title: Entry 2
```

The `list` command can be used to search for containers by passing it a
`/`-separated path:

```bash
$ rook ls Hosting
3    Accounts/Hosting   Entry 1
4    Accounts/Hosting   Entry 2
7    Internet/Hosting   Entry 1
$ rook ls Internet/Hosting
7    Internet/Hosting   Entry 1
```

This also works with the `-t` (tree) option:

```bash
$ rook ls -t Hosting
┌─ /
├─┬─ Accounts
│ └─┬─ Hosting
│   ├── Entry 1
│   └── Entry 2
└─┬─ Internet
  └─┬─ Hosting
    └── Entry 1
```

Note that `/` as the path may return unexpected results if you use `/` in your
entry names.

#### Locking and unlocking with a screensaver

If you're looking at rook, you may have a similar system, as I describe here,
all of these functions are usually provided by the DE; if you're replacing one,
you've probably replaced all of them.

On my system, I use [xss-lock](https://bitbucket.org/raymonad/xss-lock) as an
auto-locker, although
[xautolock](https://ibiblio.org/pub/Linux/X11/screensavers/) works just as well.
In both cases, the locker called is a shell script that wraps
[i3lock](https://github.com/Raymo111/i3lock-color); I do this because there are
things I want to happen when the screen locks, such as disabling dunst
messaging. In this script, you can call `rook lock --pin`, get the pin and save
it, and then use the pin to unlock rook when the screen is unlocked.

A simplified version of what I run is:

```
$ xss-lock ~/.bin/superlock
$ cat ~/.bin/superlock
##!/usr/bin/zsh

dunstctl set-paused true

lock() {
  i3lock-fancy-rapid 16 3 -enkS 1
  rook open --pin $1
}
lock $(rook lock --pin)

dunstctl set-paused false
```

or,

```
rook lock --pin | (
  i3lock-fancy-rapid 16 3 -enkS 1
  read pin && rook open --pin $pin
)
```

Just be careful about using long-term environment variables to store the pin,
because environment variables can be read from `/proc` and it increases the
attack surface. For example, I'd avoid:

```
PIN=$(rook lock --pin)
i3lock
rook open --pin $PIN
```

### Autotype

There are shell scripts in the `utils/` directory: `autotype.sh` and
`getAttr.sh`. `autotype.sh` can provide form-filling utility; it works by
grabbing the focused window's title and uses `rook match` to find the best match
in the DB -- rook uses the Keepass Autotype rules for this, if they exist -- and
the best match is returned with the first line being the defined autotype
sequence for the entry. The script then parses out the key/value pairs and uses
`xdotool` to type the sequence. `getAttr.sh` uses rofi to allow the user to
select an entry from the DB, and then an attribute from the entry, and then it
performs a user-requested action with the data.

`autotype.sh` and `getAttr.sh` are shell scripts; if you use them, you should
also read through them. They're not exactly trivial scripts, but at ~100 lines
long each they're auditable even if you're not a shell expert. Again, **I**
trust me implicitly, but you shouldn't.

`autotype.sh` is straightforward -- it just does its thing: tries to find the
best entry and type the autotype sequence. If it gets multiple matches, it shows
them in a rofi dialog, you choose one, and off it goes.

`getAttr.sh` has a `-h` argument for help, and more complex options. It can use
[fzf](https://github.com/junegunn/fzf) instead of `rofi`, and be used entirely
from the command line. It can copy the selected value to the system clipboard,
print the value out, or type the value out with `xdotool`.

`getAttr.sh` can look up an entry using the current window's title, as
`autotype.sh` does, with the `-a` argument ("autodetect"). Used with the `-u`,
`-p`, or `-o` arguments ("username", "password", "OTP"), and the `-c` argument
("copy to clipboard"), you could e.g. bind hot keys to copy values to the
clipboard which you could manually paste into the appropriate fields. This is
handy when `autotype.sh` fails (as it easily can).

It's important to recognize that `autotype.sh` is working almost blindly. It
gets, at most, a window title to try to match against records, and then it just
executes the key presses using `xdotool`. The upside is that it is not prone to
the hijacking attack DOM-based password tools (including KeePassXC itself) are
prone to. The downside is that any failure of the website to _reasonably_ focus
input fields in a reliable order can cause issues. Especially agreggious are the
sites which break up username/password into two different forms, or performm
AJAX loading, or any other messing around with the login process. In these
cases, using `getAttr.sh` is preferrable. It can still perform keypress entry,
for those _really_ hostile, obnoxious web sites which disable pasting, but it
can also copy to the clipboard. For example, with HerbstluftWM:

```bash
$ herbstclient keybind Mod4+p spawn getAttr.sh -acu
$ herbstclient keybind Mod4+u spawn getAttr.sh -acp
$ herbstclient keybind Mod4+u spawn getAttr.sh -aco
```

then logging in to, e.g., SlayPal might be a sequence:

1. Get to the SlayPal login page
2. Press `Mod4+u` to get your SlayPall username
3. Click on the "user name" field and `^v` to paste the username
4. Press `Enter` and get to the password field
5. Press `Mod4+p` to get your password
6. `^v` to paste the password, and hit `Enter`

#### autotype & getAttr Dependencies

These are dependencies **only** for the utility scripts -- the rook binary
itself does not use any of these.

- [zsh](https://www.zsh.org/), because I just couldn't be arsed to do the extra
  work to make some things in `autotype.sh` work the harder way in bash.
- [ripgrep](https://github.com/BurntSushi/ripgrep), because -- again -- it makes
  some things just so much easier.
- [xdotool](https://www.semicomplete.com/projects/xdotool/), for teh tappy-tap-tap.
- [xprop](https://gitlab.freedesktop.org/xorg/app/xprop), because Luakit needs
  hacks to work properly. If you don't use Luakit, you can (should?) delete the
  hacks -- it'll probably make things more robust.
- [yad](https://github.com/v1cont/yad), for providing info dialogs.
- [xsel](https://vergenet.net/~conrad/software/xsel/), used in `autoType.sh` to
  put values in the clipboard.

KeepassXC autotype sequences have a lot of keywords; the script understands a
small number of these -- ENTER, DELAY, TOTP, USERNAME, PASSWORD.  It should be
easy to add to this list by modifying the case switch at the end of the file --
improvement patches are welcome.

To use the script, put it in your path and bind it to a hotkey, e.g.

```bash
herbstluftwm keybind Mod4+t spawn ${HOME}/bin/autotype.sh
#### OR, if you installed with a package manager:
herbstluftwm keybind Mod4+t spawn /usr/bin/rook-autotype
```

If you installed rook with one of the binary distribution packages, and you have
the necessary required programs listed above (rofi, ripgrep, yad, etc.), then
the utility scripts will be installed in `/usr/bin/` as `rook-autotype` and
`rook-getattr`.

### Select Entry Attribute

There's a shell script in `utils/` called `getAttr.sh` which shows a list of all
entries, and when one is selected shows a list of all attributes, one of which
can be selected. If called with the `-t` argument, the selected value will be
typed using `xdotool`. If called with the `-c` arg, the value will be copied
into the clipboard using `xsel`. If no argument is given, the script prints the
value to STDOUT.

`getAttr.sh` can prompt either with `rofi` or `fzf`; the `-f` argument selects
`fzf`, and `rofi` is used by default. Using `fzf` allows `getAttr` to be used
without a GUI.

### Testing

You can see how `autotype.sh` will work by faking an input dialog with `yad`. To
get as close to real life, you'll want to run the application that will ask for
the login -- if it's Paypal, then go to Paypal.com in an incognito window and
click "Logn". If it's some GTK application, run the application and start the
login process. Then use `xprop` to get the window title -- this is what
`autotype.sh`, and incidentally, KeepassXC, uses for entry matching. Then run
`yad` with that window title as the `--title` parameter -- something like this:

```bash
yad  --form --field "Username" --field "Password:H" --field "TOTP:NUM" --title "Log in to Sourcehut"
```

Give that Username field focus and trigger your autotype -- once you hit OK, the
values will be printed on the command line.

## Install

### From your distribution

rook packages are available in Apline testing, and in Arch AUR. For Alpine,
you'll have to have the testing repo enabled:

```shell
echo "http://dl-cdn.alpinelinux.org/alpine/edge/testing/" >> /etc/apk/repositories
apk add --no-cache rook
```

For Arch, use your AUR enabled tool:

```shell
yay -Sy rook
```

### From pre-built binaries

Download the latest release from the [releases
page](https://downloads.ser1.net/software/rook/). Optionally, also download the
binary signature and validate the binary. Decompress the binary with `gzip`,
copy it somewhere in your path, and off you go.

I provide builds for Linux arm64 and amd64, and the _latest_ version as RPM,
deb, apk, and Arch zst packages for each of these.

You can get my public key either from [Sourcehut](https://meta.sr.ht/~ser.pgp),
or from one of the public key servers. You'll want to grab the one that says:

`Sean E. Russell (Signing key for Sourcehut builds) <ser@ser1.net>`, which is
key `5E0D7ABD6668FDD1`. To get it from Sourcehut, where the builds are made:

```bash
curl https://meta.sr.ht/~ser.pgp | gpg --import
####      OR    ###
gpg --keyserver keys.openpgp.org --recv-keys 6668FDD1
####  AND THEN  ###
gpg --verify rook_arm64_latest.gz.sig
```

If you install from one of the pre-built packages, the script `autotype.sh` will
be installed as `/usr/bin/rook-autotype`, and `getAttr.sh` as
`/usr/bin/rook-getattr`.

### Build from source

If you want to try building and running on something other than Linux -- say,
one of the BSDs, or Darwin -- then this or the [next section](#Install the Go
way) is what you'll need to use.

You'll need [Mercurial](https://www.mercurial-scm.org/) and
[Go](https://golang.org). For the manpages, you'll need Go 1.24 or greater.

```bash
hg clone https://hg.sr.ht/~ser/rook
go build .
## OR, to embed the correct version s.t. `rook version` works,
go build -ldflags "-X main.Version=$(hg log -r tip --template '{bookmarks}')" .
```

If you have Go 1.24 installed, you can generate the man pages with:

```
go generate
```

They'll be put in a directory called `man1`; you can then copy them into some
place in your `man --path`.

### Install the Go way

```bash
go install ser1.net/rook@latest
## OR, if you know the version you want
go install -ldflags '-X main.Version=v0.0.9' ser1.net/rook@v0.0.9
```

### nfpm

The repo contains an [`nfpm.yaml`](https://nfpm.goreleaser.com/) file, from
which can be built packages for Arch, rpm, apk, and deb systems. You could use
this to build old versions, or versions for other architectures such as
arm5/6/7. If you want to do this, you're best off reading the nfpm usage docs,
since repeating them here would be silly. But, basically, you'll need to make
sure some environment variables are set, and run (to get an apk; replacable with
`rpm`, `deb`, or `archlinux`):

```bash
VERSION=$(hg log -r tip --template '{bookmarks}') \
  GOOS=$(go env GOOS) \
  GOARCH=$(go env GOARCH) && \
  nfpm pkg -p apk
```

## Roadmap

Most of the roadmap is kept in the sourcecode. A good way of looking at it is to
get [legume](https://sr.ht/~ser/legume) and run leg in the source directory.
However, there is a general thrust to most of what's coming:

- Bug fixes
- DB modification commands. I'm bearish about adding these: I think they're best
  done through tools like KeepassXC, which provide nice UIs, and which handle
  conflicts and merging well; they also add a **lot** of complexity that would
  make auditing harder, and for questionable value. Rook is not intended to
  fully replace KeepassXC tools. That said: small changes, like changing a
  password or adding TOTP to an entry, might be handy to do on the CLI. We shall
  see.

## Credits and Related

There are countless giants that should be listed, that we all take so much for
granted that they are unknown to most of us. However, I'd like to call out a
couple of individuals whose code I've leveraged heavily, or sometimes
(basically) stolen. Those I've copied/pasted from I've also done my best to
audit in the process.

- [tobischo](https://github.com/tobischo), for whom I'm very grateful for
  allowing me to avoid reading the entire Keepass v2 spec and implement it by
  hand. Although I probably should have, in following with my own attempt to
  minimize the attack surface of rook.
- [jlinoff](https://github.com/jlinoff), whose
  [getPassword](https://gist.github.com/jlinoff/e8e26b4ffa38d379c7f1891fd174a6d0)
  (MIT) code filled a tedious hole. I very much appreciate that jlinoff not only
  provided a solution, but provided _3_ solutions with different dependencies
  ranging from stdlib only, stdlib but with some of the more esoteric packages,
  and a final one that pulls in a `golang.org/x` package.
- The rook logo was derived from SVG art titled ['Raven With Key' by Karen
  Arnold](https://openclipart.org/detail/252186/raven-with-key). The original
  was licensed CC0, and so, therefore, is this.
- So much OSS software goes into my development process. I depend, quite
  completely, on:

  - [Helix](https://helix-editor.com/) - a modern, fast editor this a vi-like
    modal paradigm and advanced features like multi-range selects.
  - [Go](https://golang.org)
  - [gopls](https://github.com/golang/tools/tree/master/gopls), the Really Very
    Good Go language server.
  - [delve](https://github.com/go-delve/delve), Go's defacto standard debugger. It
    has a GUI, even, if you're into that sort of thing.
  - [Mercurial](https://mercurial-scm.org) - still the best VCS. I spend a lot of
    time in git, and while Jujutsu makes git suck less, it's still not quite
    Mercurial.
  - [Gnupg](https://www.gnupg.org/)
  - [ripgrep](https://github.com/BurntSushi/ripgrep) - while it's popular to write
    replacements for the old POSIX toolset, ripgrep is really the only one that
    vastly improves over the original enough to switch.
  - and Linux, X, herbstluftwm, zsh, the entire standard Linux POSIX-ish
    toolkit.

  I also advocate the following tools as being Extremely Useful and Sometimes
  Indispensible™ to the development process:

  - [Sourcehut](https://sr.ht) - it's not free, but it's also not bloated with
    Javascript and CSS, and it still supports Mercurial repositories. It offers
    everything an OSS project could need - lists, tickets, a really nice CI
    system, homepage-like Project containers. It's inexpensive, especially for
    what you get, but if anything more than $0 is too expensive, it's also OSS
    -- you can download every component and host it yourself; what you're really
    paying for is the hosting.
  - [revive](https://revive.run) - a much nicer alternative to golint
  - [csvtk](https://github.com/shenwei356/csvtk) - a strange entry for a dev tool
    list, but I use this thing All. The. Time. It's part of the CI. CSV is still
    one of the best semi-structured document formats, far less bloated than JSON
    (which is already pretty light). It's indispensible in my build CI.
  - [go-licenses](https://github.com/google/go-licenses) - if you have to have dependencies, tools like this become important. It's a license jungle out there.
  - [scc](https://github.com/boyter/scc). I've used a lot of different LOC
    counters; scc is nice because it also does some complexity calculations,
    generates CSV, can disable the pretty-printing, is multi-threaded, can
    calculate DRY metrics, respects .ignore files, can identify generated and
  minified files, count duplicated LOC... it is a very versatile tool.
  - [govulncheck](https://pkg.go.dev/golang.org/x/tools/gopls/internal/vulncheck/govulncheck)
  - -
    Increasingly imortant; it doesn't cover all of the security risks, but it
    helps.
  - [markdownlinkchecker](https://github.com/becheran/mlc) - I _just_ discovered
    this when a user brought a broken link in a README to my attention. I went
    through and checked all of my projects and found a distressing number of
    broken links -- and discovered an overly-restrictive firewall rule on one of
    my servers that was preventing access to a service! So I'm currently extremely
    grateful for this tool.
  - [shields.io](https://shields.io) - provides a really valuable, free service
    for generating badges (as used at the top of this readme). I can't thank them
    enough. My CI process uses their service at build time and generates static
    versions to reduce load on their service.

rook is developed using a set of decentralized, CLI-based tools. Web tools are
fine, and they're sometimes necessary, but they centralize data, are hard to
script, and just generally get in my way. When there are many people from
different domains collaborating on a thing, this extra impediment is necessary;
when there aren't, it's not. I use code comments for issue tracking and
[legume](https://hg.sr.ht/~ser/legume) to make that easier to view; I use
[changelog](https://hg.sr.ht/~ser/changelog) to generate the
[CHANGELOG.md](https://hg.sr.ht/~ser/rook/browse/CHANGELOG.md?rev=tip).

rook was inspired by [kpmenu](https://github.com/AlessioDP/kpmenu), which
approaches security differently, requiring interaction from the user on every
request. This makes it harder to use as a secret service for automating scripts,
and since kpmenu has a very specific threat model it's been resistant to
[patches](https://git.sr.ht/~ser/kpmenu) that would weaken their solution, and
even with those patches [additional tools](https://hg.sr.ht/~ser/quasiauto) are
needed to support autotype. As mentioned in [Security](SECURITY.md), Rook assumes
the user environment is secure and can therefore take a more relaxed approach;
these are the same assumptions and approaches used by any secret service
software, such as Gnome's
[secret-tool](https://gnome.pages.gitlab.gnome.org/libsecret/),
[keyring](https://github.com/jaraco/keyring), or
[pass](https://www.passwordstore.org/).

## Other solutions

Linux users have many options for storing passwords. I can't list all of them
here, but I can list the ones I've used, and why I believe **rook** was
necessary. In every case where the tool has a bespoke data store, it means the
user having to _manually_ syncing passwords between Keepass and whatever the
tool is using, and is for this reason is IMO unsuitable.

- [KeepassXC](https://keepassxc.org). This is an outstanding program that you
  should use. It does include a _proper_ secret service option, in that it is
  compatible with keyring and Gnome's secret-tool. It is, however, a GUI tool,
  and the CLI program that comes with it requires a password every time it is
  called, making it unsuitable for background scripts.
- [kpmenu](https://github.com/AlessioDP/kpmenu). As mentioned above, kpmenu's
  architecture does not lend itself to running as a secret service that can be
  used by background jobs needing credentials.
- [secret-service](https://github.com/yousefvand/secret-service) is another
  keyring-compatible solution. It uses its own backend DB store, and I want to
  use a Keepass DB. It also has a fairly large number of external, non-stdlib
  dependencies (19), making it much harder to fully audit.
- [secret-tool](https://gnome.pages.gitlab.gnome.org/libsecret/). This is what
  Gnome uses for password stores; it has a bespoke password database (not
  Keepass), and communicates over DBus -- using secret-tool outside of Gnome
  pulls in several extra services and software dependencies, which is relatively
  heavy if a secret service is all you're using it for.
- [keyring](https://github.com/jaraco/keyring). This is (I believe) the CLI tool
  for KDE and KWallet; it is compatible with Gnome's secret service, again
  communicating over DBus. It requires the KWallet service to be running in the
  background, and IIRC that kicks off several related KDE services. Like secret-
  tool, if you're not running the KDE desktop, you're getting a chunk of the
  load of it by running `keyring`.
- [pass](https://www.passwordstore.org/). git for passwords, encrypted with GPG!
  It's a fantastic idea, and by using `gpg-agent` it's not adding much to your
  service load. It has two downsides: first, it's _still_ a bespoke backend
  password database, meaning having to manually sync any passwords or entries
  with Keepass by hand; and second, it exposes a huge amount of metadata about
  the secrets. Since the entire DB is a filesystem, where the entry titles are
  directory names and attributes are files in the directory, the entire
  structure of the database is exposed, unencrypted, and readable. Attackers may
  not know what your Pornhub password is, but they know you have an account.
  This is fine for `pass`, which is designed more for secret sharing across
  teams, and so the account names are not considered confidential information.
  It's pretty horrible for anyone else, and especially anyone considered a
  dissident for whom that metadata could be used in persecution.
- [gopass](https://github.com/gopasspw/gopass) is a Go-based client for `pass`
  databases. If you use pass, it's a pretty nice tool. `pass` itself is a
  collection of shell scripts; `gopass` is a single self-contained binary. It
  doesn't have any connection with Keepass, though.

app <- setRefClass(
    'HelloWorld',
    methods = list(
	call = function(env){
	    list(
		status=200,
		headers = list(
		    'Content-Type' = 'text/html'
		    ),
		body = paste('<h1>Hello World!</h1>')
	    )
	}
    )
)$new()

library(datasets)
library(RJSONIO)
render_json <- function(object) {

    # Some objects are instances of a sub-class of data.frame
    # and RJSONIO doesn't know what to do with them, so we just
    # use trickery.
    if (inherits(object,'data.frame',which = TRUE) > 0){

	class(object) <- 'data.frame'

	# Even these may have ts objects as columns so lets
	# just punt for right now and assign it an NA column.
	for (i in names(object)){
	    if (inherits(object[[i]],'ts')){
		object[[i]] <- NA
	    }
	}
    }

    # Unclassing here is unclassy. Would be nice to use as.data.frame
    # but it seems that the original ts object is stuffed into the result
    # somehow.
    if (inherits(object,'ts')){
	object <- unclass(object)
    }

    if (inherits(object,'table') || inherits(object,'array')){
	object <- as.data.frame(object)
    }

    RJSONIO::toJSON(object)
}

app <- Builder$new(
    Brewery$new(url='.*\\.html$',root='.'),
    URLMap$new(
	'^/.*\\.csv$' = function(env){
	    req <- Request$new(env)
	    res <- Response$new()
	    datasets <- ls('package:datasets')
	    tmpdataset <- sub('^/(.*)\\.csv','\\1',req$path_info())
	    dataset <- datasets[datasets %in% tmpdataset][1]
	    if (is.na(dataset)) dataset <- 'iris'
	    res$header('Content-type','text/csv')
	    res$write(
		paste(
		    capture.output(eval(parse(text=sprintf('write.csv(%s)',dataset)))),
		    collapse="\n"
		)
	    )
	    res$finish()
	},
	'^/.*\\.json$' = function(env){
	    req <- Request$new(env)
	    res <- Response$new()
	    datasets <- ls('package:datasets')
	    tmpdataset <- sub('^/(.*)\\.json','\\1',req$path_info())
	    dataset <- datasets[datasets %in% tmpdataset][1]
	    if (is.na(dataset)) dataset <- 'iris'
	    res$header('Content-type','application/json')
	    res$write(eval(parse(text=sprintf('render_json(%s)',dataset))))
	    res$finish()
	},
	'^/.*\\.png$' = function(env){
	    req <- Request$new(env)
	    res <- Response$new()
	    datasets <- ls('package:datasets')
	    tmpdataset <- sub('^/(.*)\\.csv','\\1',req$path_info())
	    dataset <- datasets[datasets %in% tmpdataset][1]
	    if (is.na(dataset)) dataset <- 'iris'
	    res$finish()
	},
	'.*' = Redirect$new('/index.html')
    )
)
############################################
## Top level HTML pages
############################################
#
#render_brew <- function(template, params = list(), path = getwd()) {
#	if (is.list(params)) {
#		env <- new.env(TRUE)
#			for(name in names(params)) {
#				env[[name]] <- params[[name]]
#			}
#		params <- env
#	}
#
#	path <- file.path(path, "views", stringr::str_c(template, ".html"))
#	if (!file.exists(path)) stop("Can not find ", template, " template ",
#				call. = FALSE)
#	setContentType('text/html')
#	brew::brew(path, envir = params)
#	OK
#}
#
#router$get("/", function(...) {
#	# brews the file index.html in the /views dir
#	render_brew("index",list(...))
#})
#
## This is the RESTful part of the application. Each 
## dataset found in the datasets package has its own
## URL
#router$get("/dataset/:dataset.html", function(dataset) {
#	if (any(ls('package:datasets') == dataset))
#		render_brew("dataset",list(dataset=dataset))
#	else
#		render_brew("error",list(dataset=dataset))
#})
#
#router$get("/dataset/:dataset.csv", function(dataset) {
#	setContentType('text/csv')
#	eval(parse(text=paste('write.csv(',dataset,',file=stdout())')))
#	OK
#})
#
#
#
############################################
## Web services
############################################
#
#router$get('/dataset/:dataset.json', function(dataset){
#	if (!any(ls('package:datasets') == dataset))
#		render_brew("error",list(dataset=dataset))
#	else 
#		render_json(get(dataset))
#})
#
## Returned when no dataset's example renders a plot
#bad_plot <- function(dataset){
#    t <- tempfileWithExtension()
#    png(t)
#    #par(mar=rep(0,4))
#    plot(rnorm(100),main=paste('Fail for',dataset,' but Check it!'),col=rainbow(100,alpha=runif(100,0,1)),pch='.',cex=c(2,3,4,5,10,50,100))
#    dev.off()
#	payday <- readBin(t,'raw',file.info(t)$size)
#	unlink(t)
#	payday
#}
#
#router$get('/dataset/:dataset.png', function(dataset){
#
#	if (!any(ls('package:datasets') == dataset))
#		render_brew("error",list(dataset=dataset))
#
#	setContentType('image/png')
#
#	t <- tempfileWithExtension()
#	#cat('tempfile is',t,'\n')
#	png(t)
#
#	# This is a bit of magic. R has an example function
#	# which runs example code located at the end of a
#	# particular help topic. Fortunately, there's a help
#	# topic for all datasets exported from the 'datasets'
#	# package. Unfortunately, not all of them produce a plot,
#	# and they can be noisy.
#	#
#	# This is where you would place your own data and plot routines, FYI
#	#
#	capture.output(
#	suppressWarnings(
#		eval(
#			substitute(
#				example(dataset,package='datasets',ask=FALSE),
#				list(dataset=dataset)
#			)
#		)
#	))
#
#
#	dev.off()
#	payday <- try(readBin(t,'raw',file.info(t)$size))
#	unlink(t)
#
#	if (inherits(payday,'try-error') || length(payday) <=1 ){
#		payday <- bad_plot(dataset)
#	}
#	
#	sendBin(payday)
#
#	OK
#})
#

app <- Builder$new(
    URLMap$new(
	'^/.*\\.html$' = function(env){
	    req <- Request$new(env)
	    res <- Response$new()
	    if (is.null(req$GET()$n)){
		n <- 100
	    } else {
		n <- as.integer(req$GET()$n)
	    }
	    res$write('How many squares?\n')
	    res$write('<form method="GET">\n')
	    res$write(sprintf('<input type="text" name="n" value="%d">\n',n))
	    res$write('<input type="submit" name="Submit">\n</form>\n<br>')
	    if (n>0){
		res$write(paste('<img src="',req$to_url('/plot.png',n=n),'">',sep=''))
	    }
	    res$finish()
	},
	'^/.*\\.png$' = function(env){
	    req <- Request$new(env)
	    res <- Response$new()
	    res$header('Content-type','image/png')
	    if (is.null(req$GET()$n)){
		n <- 100
	    } else {
		n <- as.integer(req$GET()$n)
	    }
	    t <- tempfile()
	    png(file=t)
	    png(t,type="cairo",width=200,height=200)
	    par(mar=rep(0,4))
	    plot(rnorm(n),col=rainbow(n,alpha=runif(n,0,1)),pch='.',cex=c(2,3,4,5,10,50))
	    dev.off()
	    res$body <- t
	    names(res$body) <- 'file'
	    res$finish()
	},
	'.*' = Redirect$new('/index.html')
    )
)

app <- function(env){
    req <- Request$new(env)
    res <- Response$new()
    getVars <- req$GET()
    postVars <- req$POST()

    content <- c("<!DOCTYPE html>\n",
      "<html>\n<head>\n<title>Rook Array Parameters</title>\n</head>\n",
      "<body>\n<h1>Rook Array Parameters</h1>\n",
      "<h2>GET request</h2>\n<form>\n<p>\n",
      "<label><input type='checkbox' name='foo[]' value='foo' /> foo</label>\n",
      "<label><input type='checkbox' name='foo[]' value='bar' /> bar</label>\n",
      "</p>\n<p>\n<input type='submit' name='get' />\n</p>\n</form>\n")

    if (length(getVars) > 0) {
      content <- c(content,
        "<pre>\n",
        paste(capture.output(str(getVars), file=NULL), collapse='\n'),
        "</pre>\n")
    }

    content <- c(content,
      "<h2>POST request</h2>\n<form method='post'>\n<p>\n",
      "<label><input type='checkbox' name='foo[]' value='foo' /> foo</label>\n",
      "<label><input type='checkbox' name='foo[]' value='bar' /> bar</label>\n",
      "</p>\n<p>\n<input type='submit' name='post' />\n</p>\n</form>\n")

    if (length(postVars) > 0 && 'post' %in% names(postVars)) {
      content <- c(content,
        "<pre>\n",
        paste(capture.output(str(postVars), file=NULL), collapse='\n'),
        "</pre>\n")
    }

    content <- c(content,
      "<h2>POST multipart request</h2>\n<form method='post' enctype='multipart/form-data'>\n<p>\n",
      "<label><input type='checkbox' name='foo[]' value='foo' /> foo</label>\n",
      "<label><input type='checkbox' name='foo[]' value='bar' /> bar</label>\n",
      "</p>\n<p>\n<input type='submit' name='multipart' />\n</p>\n</form>\n")

    if (length(postVars) > 0 && 'multipart' %in% names(postVars)) {
      content <- c(content,
        "<pre>\n",
        paste(capture.output(str(postVars), file=NULL), collapse='\n'),
        "</pre>\n")
    }

    res$write(c(content, "</body>\n</html>\n"))
    res$finish()
}

app <- function(env){
    req <- Request$new(env)
    res <- Response$new()
    envstr <- paste(capture.output(str(as.list(env)),file=NULL),collapse='\n')
    poststr <- paste(capture.output(str(req$POST()),file=NULL),collapse='\n')
    getstr <- paste(capture.output(str(req$GET()),file=NULL),collapse='\n')
    randomString <- function() paste(letters[floor(runif(10,0,26))],collapse='')
    randomNumber <- function() runif(1,0,26)
    res$write( c(
    '<HTML><head><style type="text/css">\n',
    'table { border: 1px solid #8897be; border-spacing: 0px; font-size: 10pt; }',
    'td { border-bottom:1px solid #d9d9d9; border-left:1px solid #d9d9d9; border-spacing: 0px; padding: 3px 8px; }',
    'td.l { font-weight: bold; width: 10%; }\n',
    'tr.e { background-color: #eeeeee; border-spacing: 0px; }\n',
    'tr.o { background-color: #ffffff; border-spacing: 0px; }\n',
    '</style></head><BODY>',
    '<img alt="rook logo" src="http://wiki.rapache.net/static/rook.png">',
    '<H1>Welcome to Rook</H1>\n',
    sprintf('<form enctype="multipart/form-data" method=POST action="%s/%s?called=%s">',env$SCRIPT_NAME,randomString(),randomNumber()),
    'Enter a string: <input type=text name=name value=""><br>\n',
    'Enter another string: <input type=text name=name2 value=""><br>\n',
    'Upload a file: <input type=file name=fileUpload><br>\n',
    'Upload another file: <input type=file name=anotherFile><br>\n',
    '<input type=submit name=Submit><br><br>',
    'Environment:<br><pre>',envstr,'</pre><br>',
    'Get:<br><pre>',getstr,'</pre><br>',
    'Post:<br><pre>',poststr, '</pre><br><br>'
    ))
    res$finish()
}

library(Hmisc)
dir.create(file.path(tempdir(),'plots'),showWarnings=FALSE)
app <- Builder$new(
    Static$new(
	urls = c('/css','/images','/javascript'),
	root = '.'
    ),
    Static$new(urls='/plots',root=tempdir()),
    Brewery$new(
	url='/brew',
	root='.',
	imagepath=file.path(tempdir(),'plots'),
	imageurl='../plots/'
    ),
    Redirect$new('/brew/useR2007.rhtml')
)

app <- function(env) {
    req <- Rook::Request$new(env)
    res <- Rook::Response$new()
    res$write('Choose a CSV file:\n')
    res$write('<form method="POST" enctype="multipart/form-data">\n')
    res$write('<input type="file" name="data">\n')
    res$write('<input type="submit" name="Upload">\n</form>\n<br>')

    if (!is.null(req$POST())){
	data <- req$POST()[['data']]
	res$write("<h3>Summary of Data</h3>");
	res$write("<pre>")
	res$write(paste(capture.output(summary(read.csv(data$tempfile,stringsAsFactors=FALSE)),file=NULL),collapse='\n'))
	res$write("</pre>")
	res$write("<h3>First few lines (head())</h3>");
	res$write("<pre>")
	res$write(paste(capture.output(head(read.csv(data$tempfile,stringsAsFactors=FALSE)),file=NULL),collapse='\n'))
	res$write("</pre>")
    }
    res$finish()
}

app <- function(env){
    req <- Rook::Request$new(env)
    res <- Rook::Response$new()
    friend <- 'World'
    if (!is.null(req$GET()[['friend']]))
	friend <- req$GET()[['friend']]
    res$write(paste('<h1>Hello',friend,'</h1>\n'))
    res$write('What is your name?\n')
    res$write('<form method="GET">\n')
    res$write('<input type="text" name="friend">\n')
    res$write('<input type="submit" name="Submit">\n</form>\n<br>')
    res$finish()
}

library(Rook)
library(RJSONIO)
app <- function(env){
    req <- Request$new(env)
    res <- Response$new()

    obj <- sub('^/','',req$path_info())

    
    # Get out of here fast if no object exists
    if (!exists(obj)) return(res$finish())

#    res$write(paste(capture.output(str(req$params()),file=NULL),collapse='\n'))
#    return(res$finish())

    # Gather args from one of three sources: GET, POST as a x-www-urlencoded,
    # or POST as JSON payload. params() squishes GET and POST together when
    # POST is x-www-urlencoded.
    if (!is.null(req$params())) {
	args <- req$params()
    } else {
	# TODO: Collect POST payload and pass to RJSONIO
	args <- list()
    }

    # Normalize arguments to R types if necessary. Integers to integer, Numerics to numeric, etc.
    # Maybe we can propose a vector syntax in CGI that's coherent, too.
    for (i in names(args)){

	# Keep as character anything that starts with a quote char
	if (grepl('^[\'"]',args[[i]]))
	    next

	# Integer
	if (grepl('^[+-]?\\d+$',args[[i]]))
	    args[[i]] <- as.integer(args[[i]])

	# Numeric. need to add scientific notation
	if (grepl('^[+-]?[0-9.]+$',args[[i]]))
	    args[[i]] <- as.numeric(args[[i]])
    }
    
    if (is.function(get(obj)))
	res$write(toJSON(do.call(obj,args)))
    else
	res$write(toJSON(get(obj)))
	
    res$finish()
}

.rApacheInputStream <- setRefClass(
   'rApacheInputStream',
   methods = list(
      read_lines = function(n = -1L){
         if (n<=0) return(character())
         readLines(n=n,warn=FALSE)
      },
      read = function(l = -1L){
         receiveBin(l)
      },
      rewind = function(){
         warning("rApache doesn't support rewind()")
      }
   )
)

.rApacheErrorStream <- setRefClass(
   'rApacheErrorStream',
   methods = list(
      flush = function() { base::flush(stderr()) },
      cat = function(...,sep=" ",fill=FALSE,labels=NULL)
      { base::cat(...,sep=sep,fill=fill,labels=labels,file=stderr()) }
   )
)

Server <- setRefClass(
   'rApacheServer',
   fields = c('appPath','appList'),
   methods = list(
      initialize = function(...){
         callSuper(...)
      },
      AppPath = function(appPath){
         if (length(appList) == 0) return()
         appPath <<- appPath
      },
      build_env = function(){
         env <- new.env(hash=TRUE,parent=emptyenv())

         lapply(names(SERVER$headers_in),function(h){
               assign(
                  paste('HTTP_',gsub('-','_',gsub('(\\w+)','\\U\\1',h,perl=TRUE)),sep=''),
                  SERVER$headers_in[[h]],
                  env)
         })

         if (exists('HTTP_CONTENT_LENGTH',env))
            assign('CONTENT_LENGTH',get('HTTP_CONTENT_LENGTH',env),env)
         else 
            assign('CONTENT_LENGTH',SERVER$clength,env)

         if (exists('HTTP_CONTENT_TYPE',env))
            assign('CONTENT_TYPE',get('HTTP_CONTENT_TYPE',env),env)
         else
            assign('CONTENT_TYPE',SERVER$content_type,env)

         assign('SCRIPT_NAME',SERVER$cmd_path,env)

         assign('PATH_INFO',sub(SERVER$cmd_path,'',SERVER$uri),env)

         # Ensure only one leading forward slash
         env$PATH_INFO <- sub('^/+','/',env$PATH_INFO)

         assign('QUERY_STRING',SERVER$args,env)
         assign('QUERY_STRING',ifelse(is.null(SERVER$args),'',SERVER$args),env)
         assign('REQUEST_METHOD',SERVER$method,env)

         assign('REMOTE_HOST',SERVER$remote_host,env)
         assign('REMOTE_ADDR',SERVER$remote_ip,env)

         hostport <- strsplit(get('HTTP_HOST',env),':',fixed=TRUE)[[1]]

         assign('SERVER_NAME',hostport[1],env)
         if ('port' %in% names(SERVER))
            assign('SERVER_PORT',SERVER$port,env)
         else
            assign('SERVER_PORT',hostport[2],env)

         assign('rook.version',packageDescription('Rook',fields='Version'),env)
         assign('rook.url_scheme', ifelse(isTRUE(SERVER$HTTPS),'https','http'),env)
         assign('rook.input',.rApacheInputStream$new(),env)
         assign('rook.errors',.rApacheErrorStream$new(),env)

         env
      },
      call = function(app){
         if (is(app,'refClass')) res <- try(app$call(build_env()))
         else if (is(app,'function')) res <- try(app(build_env()))
         else stop('App not Rook aware')

         if (inherits(res,'try-error')){
            warning('App returned try-error object')
            return(HTTP_INTERNAL_SERVER_ERROR)
         }

         setStatus(res$status)

         setContentType(res$headers$`Content-Type`)
         res$headers$`Content-Type` <- NULL
         lapply(names(res$headers),function(n)setHeader(n,res$headers[[n]]))

         # If body is named, then better be a file.
         if (!is.null(names(res$body)) && names(res$body)[1] == 'file'){
            sendBin(readBin(res$body[1],'raw',n=file.info(res$body[1])$size))
         } else {
            if ((is.character(res$body) && nchar(res$body)>0) || 
                (is.raw(res$body) && length(res$body)>0) )
            sendBin(res$body)
         }

         OK
      }
   )
)$new()

Request$methods(
   GET = function(){
      if (!exists('rook.request.query_list',env))
         env[['rook.request.query_list']] <<- base::get('GET','rapache')
      env[['rook.request.query_list']]
   },
   POST = function() {
      if (exists('rook.request.form_list',env))
         return(env[['rook.request.form_list']])
      postvar <- base::get('POST','rapache')
      filevar <- base::get('FILES','rapache')

      if (length(postvar) <= 0 && length(filevar) <= 0) return(NULL)

      if (length(filevar) > 0){
         if (length(postvar) <= 0) postvar <- list()
         for (n in names(filevar)){
            if (length(filevar[[n]])>0){
               postvar[[n]] <- list(
                  filename = filevar[[n]]$name,
                  tempfile = filevar[[n]]$tmp_name,
                  content_type = Mime$mime_type(Mime$file_extname(filevar[[n]]$name))
               )
            }
         }
      }
      for (n in names(postvar)){
         if (is.null(postvar[[n]])) postvar[[n]] <- NULL
      }
      env[['rook.request.form_list']] <<- postvar
      postvar
   }
)

Redirect <- setRefClass(
    'Redirect',
    fields = c('url'),
    methods = list(
	initialize = function(url,...){
	    url <<- url
	    callSuper(...)
	},
	call = function(env){
	    req <- Rook::Request$new(env)
	    res <- Rook::Response$new()
	    res$redirect(req$to_url(url))
	    res$finish()
	}
    )
)

# Server object available to web servers to set how they please.
# Must be set in onLoad. After that, they are locked.
Server <- NULL

.onLoad <- function(libpath, pkgname){
    if ('(embedding)' %in% names(getLoadedDLLs()) && 'rapache' %in% search()){
	sys.source(
	    file.path(libpath,pkgname,'servers','rApache.R'),
	    envir = asNamespace('Rook')
	)
    }
}

RhttpdApp <- setRefClass(
    'RhttpdApp',
    fields = c('app','name','path','appEnv'),
    methods = list(
       initialize = function(app=NULL,name=NULL,...){
          if (is.null(name) || !is.character(name))
             base::stop("Need a proper app 'name'")

          .self$name <- name

          if (is.character(app) && file.exists(app)){
             oldwd <- setwd(dirname(app))
             on.exit(setwd(oldwd))
             appEnv <<- new.env(parent=globalenv())
             appEnv$.appFile <<- normalizePath(basename(app))
             appEnv$.mtime <<- as.integer(file.info(appEnv$.appFile)$mtime)
             sys.source(appEnv$.appFile,envir=.self$appEnv)

             if (exists(.self$name,.self$appEnv,inherits=FALSE))
                .self$app <- get(.self$name,.self$appEnv)
             else if (exists('app',.self$appEnv,inherits=FALSE))
                .self$app <- get('app',.self$appEnv)
             else
                base::stop("Cannot find a suitable app in file ",app)

          } else {
             .self$app <- app
             .self$name <- name
             .self$appEnv <- NULL
          }

          if (!is_rookable(.self$app))
             base::stop("App is not rookable'")

          .self$path <- ifelse(.self$name=='httpd','', paste('/custom',.self$name,sep='/'))
          callSuper(...)
       }
   )
)

RhttpdInputStream <- setRefClass(
    'RhttpdInputStream',
    fields = c('postBody','pos'),
    methods = list(
	initialize = function(postBody=NULL,...){
	    if (is.null(postBody)) {
		postBody <<- raw(0) # empty post body
	    } else if (is.character(postBody)){
		postBody <<- charToRaw(paste(names(postBody),postBody,sep='=',collapse='&'))
	    } else {
		postBody <<- postBody
	    }
	    pos <<- 1
	    callSuper(...)
	},
	read_lines = function(n = -1L){
	    if (n==0 || pos > length(postBody)) return(character())
	    nls <- which(charToRaw('\n')==postBody[pos:length(postBody)])
	    rv <- character(ifelse(n>-1L,n,0))
	    lc <- 1
	    for (i in nls){
		rv[lc] <- rawToChar(postBody[pos:(pos+i-1)])
		pos <<- pos + pos + i
		lc <- lc + 1
		if (pos > length(postBody) || lc > n)
		    break
	    }
	    if (lc <= n && pos <= length(postBody))
		rv[lc] <- rawToChar(postBody[pos:length(postBody)])
	    rv
	},
	read = function(l = -1L){
	    if (l == 0 || pos >= length(postBody)) return(raw())
	    if (l < 0){
		rv <- postBody[pos:length(postBody)]
		pos <<- length(postBody)
		return(rv)
	    } else {
		rv <- postBody[pos:(pos+(l-1))]
		pos <<- pos + l
		return(rv)
	    }
	},
	rewind = function(){
	    pos <<- 1
	}
    )
)

RhttpdErrorStream <- setRefClass(
    'RhttpdErrorStream',
    methods = list(
	flush = function() { base::flush(stderr()) },
	cat = function(...,sep=" ",fill=FALSE,labels=NULL)
	{ base::cat(...,sep=sep,fill=fill,labels=labels,file=stderr()) }
    )
)

Rhttpd <- setRefClass(
    'Rhttpd',
    fields = c('appList','listenAddr','httpdOrig','listenPort'),
    methods = list(
	initialize = function(...){
	    appList <<- list()
	    listenAddr <<- '127.0.0.1'
	    listenPort <<- 0L
	    callSuper(...)
	},
	finalize = function(){
	    if (length(appList) == 0) return()
	    for (i in rev(1:length(appList))){
		remove(appList[[i]])
	    }
	},
	full_url = function(i){
	    paste('http://',listenAddr,':',listenPort,appList[[i]]$path,sep='')
	},
	launch = function(...){
	    .self$start(quiet=TRUE)		
	    # Try to create a new app from the supplied arguments
	    app <- RhttpdApp$new(...)
	    if (add(app)){
		appName <- app$name
		browseURL(full_url(which(appName == names(appList))))
		invisible()
	    } else {
		base::stop("No app to launch")
	    }
	},
	open = function(x){
	    if (missing(x)) return(print())
	    if (is.numeric(x) || is.integer(x)){
		x <- as.integer(x)
		if (!is.null(appList[[x]]))
		    return(invisible(browseURL(full_url(x))))
		else
		    base::stop("No app at index ",x)
	    } else if (is.character(x)){
		for (i in 1:length(appList)){
		    if (appList[[i]]$name==x){
			return(invisible(browseURL(full_url(x))))
		    }
		}
		base::stop("No app named",x)
	    }
	    base::stop("Argument must be an integer or character")
	},
	browse = function(x) open(x),
   start = function(listen='127.0.0.1',port=getOption('help.ports'),quiet=FALSE){

      if(nzchar(Sys.getenv("R_DISABLE_HTTPD"))) {
         warning("httpd server disabled by R_DISABLE_HTTPD", immediate. = TRUE)
         utils::flush.console()
         return(invisible())
      }

      if(grepl('rstudio',base::.Platform$GUI,ignore.case=TRUE)){
         # RStudio has already set up host and port
         listenPort <<- tools:::httpdPort
         if (!missing(port))
            warning("RStudio has already started the web server on port ",tools:::httpdPort)
         return(invisible())
      }

      if (!missing(listen) && listen != '127.0.0.1'){
         listen <- '127.0.0.1'
         warning("This version of Rook can only listen on the loopback device.");
      }

      if (!missing(port)){
         oldPorts <- getOption('help.ports')
         on.exit(options(help.ports=oldPorts))
         options(help.ports=port)
      }

      if (length(appList) == 0)
         add(RhttpdApp$new(system.file('exampleApps/RookTestApp.R',package='Rook'),name='RookTest'))


      listenPort <<- startDynamicHelp(TRUE)

      if (listenPort == 0){
         base::stop("The internal web server could not be started!")
      }

      if (!quiet){
         cat('\nServer started on host',listen,'and port',listenPort,'. App urls are:\n\n')
         invisible(lapply(names(appList),function(i){
            cat('\thttp://',listen,':',listenPort,appList[[i]]$path,'\n',sep='')
               }))
      }
      invisible()
   },

	stop = function(){
	    listenPort <<- startDynamicHelp(FALSE)
	},
	add = function(app=NULL,name=NULL){

	    if (!inherits(app,'RhttpdApp'))
		app <- RhttpdApp$new(name=name,app=app)
	    if (!inherits(app,'RhttpdApp'))
		base::stop("Need an RhttpdApp object")

	    appList[[app$name]] <<- app
	    if(app$name=='httpd'){
		base::stop("Rook no longer supports assignment to tools:::httpd")
		#.self$httpdOrig <- tools:::httpd
		#assignInNamespace(
		#    app$name,
		#    function(path,query,postBody,headers) 
		#	.self$handler(app$name,path,query,postBody,headers), 
		#    'tools'
		#)
	    } else {
		assign(
		    app$name, 
		    function(path,query,postBody,headers) 
			.self$handler(app$name,path,query,postBody,headers), 
		    tools:::.httpd.handlers.env
		)
	    }

	    invisible(TRUE)
	},
	remove = function(app=NULL,all=FALSE){
	    if (all==TRUE){
		lapply(names(appList),remove)
		return(invisible(TRUE))
	    }
	    if (inherits(app,'RhttpdApp'))
		name <- app$name
	    else if (is.character(app))
		name <- app
	    else if (is.numeric(app) || is.integer(app))
		name <- appList[[app]]$name
	    else
		base::stop("Can only remove by object, app name, or index.")

	    if (is.null(appList[[name]])) return(FALSE)

	    appList[[name]] <<- NULL
	    ret <- FALSE
	    if(name=='httpd' && !is.null(httpdOrig)){
		tools:::httpd <- httpdOrig
		ret <- TRUE
	    } else if (exists(name,tools:::.httpd.handlers.env)){
		rm(list=name,pos=tools:::.httpd.handlers.env)
		ret <- TRUE
	    }
	    invisible(ret)
	},
	parse_headers = function(headers,env){

	    hlines <- strsplit(rawToChar(headers),'\n')[[1]]

	    lapply(
		strsplit(hlines,': '),
		function(x) {
		    assign(
			paste('HTTP_',gsub('-','_',gsub('(\\w+)','\\U\\1',x[1],perl=TRUE)),sep=''),
			x[2],
			env
			)
		}
	    )
	},
   build_env = function(appPath,path,query,postBody,headers){
      env <- new.env(hash=TRUE,parent=emptyenv())

      parse_headers(headers,env)

      # remove HTTP_ from content length and type
      if (exists('HTTP_CONTENT_LENGTH',env) && exists('HTTP_CONTENT_TYPE',env)){
         assign('CONTENT_LENGTH',env$`HTTP_CONTENT_LENGTH`,env)
         assign('CONTENT_TYPE',env$`HTTP_CONTENT_TYPE`,env)
         rm('HTTP_CONTENT_LENGTH','HTTP_CONTENT_TYPE',envir=env)
      }

      assign('SCRIPT_NAME',appPath,env)
      assign('PATH_INFO',sub(appPath,'',path,fixed=TRUE),env)

      # The R internal web server unescapes the query, so in order
      # abid the Rook spec, we have to do things in reverse:
      #
      # escape the query object so that subsequent URI building
      # methods will add the correct query string.
      assign('QUERY_STRING',
         ifelse(is.null(query),
            '',
            paste(names(query),Utils$escape(query),sep='=',collapse='&')
            ),
         env
         )
      
      if(exists("HTTP_REQUEST_METHOD", env)){
        assign('REQUEST_METHOD',get("HTTP_REQUEST_METHOD", env) ,env)
      } else {      
        assign('REQUEST_METHOD',ifelse(is.null(postBody),'GET','POST'),env)
      }

      hostport <- strsplit(get('HTTP_HOST',env),':',fixed=TRUE)[[1]]

      assign('SERVER_NAME',hostport[1],env)
      assign('SERVER_PORT',hostport[2],env)

      assign('rook.version',packageDescription('Rook',fields='Version'),env)
      assign('rook.url_scheme','http',env)
      assign('rook.input',RhttpdInputStream$new(postBody),env)
      assign('rook.errors',RhttpdErrorStream$new(),env)

      if (debug()>1)
         str(as.list(env))
      env
   },
   handler = function(appName,path,query,postBody,headers){
      if (debug()>0){
         cat('Request:',path,'\n')
      }
      app <- appList[[appName]]
      if (is.null(app)){
         base::stop("No app installed named ",appName)
         return()
      }
      if (!is.null(app$appEnv)){
         file_path = app$appEnv$.appFile
         mtime <- as.integer(file.info(file_path)$mtime)


         if (mtime > app$appEnv$.mtime){
            add(name=appName,app=file_path)
         }
         app <- appList[[appName]]
         if (is.null(app)){
            stop("No app installed named ",appName)
            return()
         }

         oldwd <- setwd(dirname(app$appEnv$.appFile))
         on.exit(setwd(oldwd))
      }
      env <- build_env(app$path,path,query,postBody,headers)
      if (is(app$app,'function')) {
         res <- try(app$app(env))
      } else {
         res <- try(app$app$call(env))
      }
      if (inherits(res,'try-error') || (is.character(res) && length(res) == 1))
         res
      else {
         # Only need to handle the case where body is a vector of strings
         # We presume that if res$body is a location to a file then the
         # app has so named it. We also presume that res$body may be
         # a raw vector, but we let the internal web server deal with that.
         if (is.character(res$body) && length(res$body) > 1){
            res$body <- paste(res$body,collapse='')
         }
         contentType <- res$headers$`Content-Type`;
         res$headers$`Content-Type` <- NULL;

         # The internal web server expects a list like the below,
         # and the position of each element is important.
         ret <- list(
            payload = res$body,
            `content-type` = contentType,
            headers = NULL,
            `status code` = res$status
         )

         # Change the name of payload to file in the case that
         # payload *is* a filename
         if (!is.null(names(res$body)) && names(res$body)[1] == 'file'){
            names(ret) <- c('file',names(ret)[-1])
         }

         # Rhttpd doesn't allow Content-Length in the headers, so delete
         # it as well
         res$headers$`Content-Length` <- NULL;
         if (length(res$headers)>0){
            ret$headers <- paste(names(res$headers),': ',res$headers,sep='')
         }

         if (debug()>0){
            cat('Response:\n')
            str(ret)
         }
         ret
      }
   },
	print = function() {
	    if (listenPort > 0){
		cat("Server started on ",listenAddr,":",listenPort,"\n",sep='')
	    } else {
		cat("Server stopped\n")
	    }
	    if (length(appList) == 0){
		cat("No applications installed\n")
		return(invisible())
	    }
	    len <- max(nchar(names(appList)))
	    for (i in 1:length(appList)){
		appName <- sprintf(paste('%-',len,'s',sep=''),names(appList)[i])
		cat('[',i,'] ',appName,' ',full_url(i),'\n',sep='')
	    }
	    cat("\nCall browse() with an index number or name to run an application.\n")
	    invisible()
	},
	show = function() print(),
	debug = function(){
	    d <- getOption('Rhttpd_debug')
	    if (!is.null(d)) as.integer(d)
	    else 0
	}
    )
)

File <- setRefClass(
    'File',
    fields = c('root','path_info','path'),
    methods = list(
	initialize  = function(root,...){
	    root <<- root
	    callSuper(...)
	},
	call = function(env){
	    path_info <<- Utils$unescape(env[["PATH_INFO"]])

	    if (length(grep('..',path_info,fixed=TRUE))){
			return(forbidden())
		}

      if (grepl('#',path_info))
        path_info <<- strsplit(path_info,'#')[[1]]

      if (grepl('\\?',path_info))
        path_info <<- strsplit(path_info,'\\?',)[[1]]

	    path <<- normalizePath(file.path(root,path_info))

	    if (file_test('-d',path)){
        if(!grepl(".*/$", path_info)){
          return(redirect(paste(env[["SCRIPT_NAME"]], env[["PATH_INFO"]], "/", sep=""), status=301))
        }
			newpath <- file.path(path, "index.html")
			if(file.exists(newpath)){
				path <<- normalizePath(newpath)
				serving()
			} else {
				return(indexdir())
			}
		} else if (file.exists(path)){
			serving()
		} else {
			not_found()
		}
	},
	forbidden = function(){
	    body = 'Forbidden\n'
	    list(
		status=403L,
		headers = list(
		    'Content-type' = 'text/plain',
		    'Content-Length'  = as.character(nchar(body)),
		    'X-Cascade' = 'pass'
		),
		body = body
	    )
	},
	indexdir = function(){
    body <- paste(list.files(path), collapse="\n")
		list(
			status=200L,
			headers = list(
				'Content-type' = 'text/plain',
				'Content-Length'  = as.character(nchar(body))
			),
			body = body
		)		
	},
  redirect = function(location){
    res <- Response$new()
    res$redirect(location, status=302)
    res$finish()    
  },
	serving = function(){
	    fi <- file.info(path)
	    if (fi$size > 0) {
		body = readBin(path,'raw',fi$size)
	    } else {
		body <- path
		names(body) <- 'file'
	    }
	    list (
		status=200L,
		headers = list(
		    'Last-Modified' = Utils$rfc2822(fi$mtime),
		    'Content-Type' = Mime$mime_type(Mime$file_extname(basename(path))),
		    'Content-Length' = as.character(fi$size)
		),
		body=body
	    )
	},
	not_found = function(){
	    body <- paste("File not found:",path_info,"\n")
	    list(
		status=404L, 
		headers = list(
		    "Content-Type" = "text/plain",
		    "Content-Length" = as.character(nchar(body)),
		    "X-Cascade" = "pass"
		),
		body = body
	    )
	}
    )
)

.RookeryApp <- setRefClass(
   'RookeryApp',
   fields = c('app','name','appEnv','configured','workingDir'),
   methods = list(
      initialize = function(app=NULL,name=NULL,...){
         if (is.null(name) || !is.character(name)){
            base::warning("Need a proper app 'name'")
            .self$configured <- FALSE
            return(callSuper(...))
         }

         .self$name <- name

         if (is.character(app) && file.exists(app)){
            .self$workingDir <- dirname(app)
            oldwd <- setwd(.self$workingDir)
            on.exit(setwd(oldwd))
            appEnv <<- new.env(parent=globalenv())
            appEnv$.appFile <<- normalizePath(basename(app))
            appEnv$.mtime <<- as.integer(file.info(appEnv$.appFile)$mtime)
            sys.source(appEnv$.appFile,envir=.self$appEnv)

            if (exists(.self$name,.self$appEnv,inherits=FALSE))
               .self$app <- get(.self$name,.self$appEnv)
            else if (exists('app',.self$appEnv,inherits=FALSE))
               .self$app <- get('app',.self$appEnv)
            else {
               base::warning("Cannot find a suitable app in file ",app)
               .self$app <- NULL
            }
         } else {
            base::warning("File does not exist: ",app)
            .self$configured <- FALSE;
         }

         if (!is_rookable(.self$app)){
            base::warning("App ",name," is not rookable'")
            .self$configured <- FALSE;
         } else {
            .self$configured <- TRUE;
         }

         callSuper(...)
      }
   )
)

.Rookery <- setRefClass(
   'Rookery',
   fields = c('req','res','appHash','messages'),
   methods = list(
      initialize = function(...){
         appHash <<- new.env()
         messages <<- list(
            emptypath = c(
               '<h2>Oops! option Rook.Rookery.paths is NULL</h2>',
               '<p>You must set this to a character vector containing',
               'valid directories where Rook apps live.</p>'
            ),
            nodots = c(
               '<h2>Apps cannot be named . or ..</h2>'
            )
         )
         callSuper(...)
      },
      message = function(name,opt=NULL){
         msg <- paste(messages,collapse='\n')
         if (!is.null(opt))
            msg <- sprintf(msg,opt)
         res$header('Content-Type','text/html')
         res$write(msg)
      },
      findSuitableApp = function(appName){
         if (appName %in% c('.','..')){
            message('nodots')
            return(NULL)
         }
         paths <- getOption('Rook.Rookery.paths')

         if (is.null(paths)){
            message('emptypath')
            return(NULL)
         }
         #for (p in paths){
         #   appReg <- paste('^',appName,'$',sep='')
         #   if (any(grepl(appReg,basename(list.dirs(p,recursive=FALSE))))){
         #   } else if (
         #}
      },
      listAllApps = function(){
      },
      call = function(env){
         req <<- Request$new(env)
         res <<- Response$new()

         # Captures foo from "/foo/.*". Presumes leading /.
         appName <- strsplit(req$path_info(),'/',fixed=TRUE)[[1]][2]
         if (is.na(appName)){
            listAllApps()
         } else {
            app <- findSuitableApp(appName)

            if (!is.null(app)){
               new_path_info <- req$path_info()
               req$path_info(sub(paste("/",appName,sep=''),'',new_path_info))
               oldwd <- setwd(app$workingDir)
               on.exit(setwd(oldwd))
               if (is(app$app,'function')) {
                  return(app$app(env))
               } else {
                  return(app$app$call(env))
               }
            }
         }

         #res$write("<pre>")
         #res$write(paste(capture.output(ls.str(env)),collapse='\n'))
         #res$write("</pre>")

         res$finish()
      }
   )
)

Static <- setRefClass(
    'Static',
    'contains' = 'Middleware',
    fields = c('urls','file_server'),
    methods = list(
	initialize = function(urls=c(),root=pwd(),...){
	    urls <<- paste('^',urls,sep='')
	    file_server <<- Rook::File$new(root)
	    callSuper(...)
	},
	call = function(env){
	    path <- env[["PATH_INFO"]]
	    if (any(sapply(urls,function(i)length(grep(i,path))>0)))
		file_server$call(env)
	    else
		app$call(env)
	}
    )
)

Response <- setRefClass(
   'Response',
   fields = c('body','status','headers','length'),
   methods = list(
      initialize = function(body='',status=200,headers=list(),...){
         .self$status <- as.integer(status)
         .self$headers <- as.environment(list('Content-Type'='text/html'))
         if (length(headers) > 0)
            .self$headers <- as.environment(c(as.list(.self$headers),headers))
         if (!is.character(body) && !is.raw(body)){
            base::stop('Body must be a character or raw vector, not',typeof(body))
         }
         .self$body <- body 
         .self$length <- Utils$bytesize(.self$body)
         callSuper(...)
      },
      header = function(key,value) {
         if (missing(value)) headers[[key]]
         else headers[[key]] <<- value
      },
      set_cookie = function(key,value){
         Utils$set_cookie_header(headers,key,value)
      },
      delete_cookie = function(key,value){
         Utils$delete_cookie_header(headers, key, value)
      },
      redirect = function(target,status=302){
         status <<- as.integer(status)
         header('Location',target)
      },
      finish = function(){
         list(
            status=status,
            headers = as.list(headers),
            body = body
         )
      },
      write = function(str){
         s <- paste(as.character(str),collapse='')
         length <<- length + Utils$bytesize(s)
         header('Content-Length',length)
         body <<- paste(body,s,sep='')
      }

   )
)

Utils <- setRefClass(
    'Utils',
    fields = c(
	'HTTP_STATUS_CODES',
	'STATUS_WITH_NO_ENTITY_BODY',
	'CHAR_TO_STATUS_CODE',
	'DEFAULT_SEP'
    ),
    methods = list(
	initialize = function(...){
	    HTTP_STATUS_CODES <<- new.env(hash=TRUE)
	    # Generated with:
	    #   curl -s http://www.iana.org/assignments/http-status-codes | \
	    #     ruby -ane 'm = /^(\d{3}) +(\S[^\[(]+)/.match($_) and
	    #                puts "      `#{m[1]}`  <- \x27#{m[2].strip}\x27"'
	    with(HTTP_STATUS_CODES,{
		`100`  <- 'Continue'
		`101`  <- 'Switching Protocols'
		`102`  <- 'Processing'
		`200`  <- 'OK'
		`201`  <- 'Created'
		`202`  <- 'Accepted'
		`203`  <- 'Non-Authoritative Information'
		`204`  <- 'No Content'
		`205`  <- 'Reset Content'
		`206`  <- 'Partial Content'
		`207`  <- 'Multi-Status'
		`208`  <- 'Already Reported'
		`226`  <- 'IM Used'
		`300`  <- 'Multiple Choices'
		`301`  <- 'Moved Permanently'
		`302`  <- 'Found'
		`303`  <- 'See Other'
		`304`  <- 'Not Modified'
		`305`  <- 'Use Proxy'
		`306`  <- 'Reserved'
		`307`  <- 'Temporary Redirect'
		`400`  <- 'Bad Request'
		`401`  <- 'Unauthorized'
		`402`  <- 'Payment Required'
		`403`  <- 'Forbidden'
		`404`  <- 'Not Found'
		`405`  <- 'Method Not Allowed'
		`406`  <- 'Not Acceptable'
		`407`  <- 'Proxy Authentication Required'
		`408`  <- 'Request Timeout'
		`409`  <- 'Conflict'
		`410`  <- 'Gone'
		`411`  <- 'Length Required'
		`412`  <- 'Precondition Failed'
		`413`  <- 'Request Entity Too Large'
		`414`  <- 'Request-URI Too Long'
		`415`  <- 'Unsupported Media Type'
		`416`  <- 'Requested Range Not Satisfiable'
		`417`  <- 'Expectation Failed'
		`422`  <- 'Unprocessable Entity'
		`423`  <- 'Locked'
		`424`  <- 'Failed Dependency'
		`425`  <- 'Reserved for WebDAV advanced'
		`426`  <- 'Upgrade Required'
		`500`  <- 'Internal Server Error'
		`501`  <- 'Not Implemented'
		`502`  <- 'Bad Gateway'
		`503`  <- 'Service Unavailable'
		`504`  <- 'Gateway Timeout'
		`505`  <- 'HTTP Version Not Supported'
		`506`  <- 'Variant Also Negotiates'
		`507`  <- 'Insufficient Storage'
		`508`  <- 'Loop Detected'
		`509`  <- 'Unassigned'
		`510`  <- 'Not Extended'
	    })

	    STATUS_WITH_NO_ENTITY_BODY <<- as.integer(c(100:199,204,304))

	    CHAR_TO_STATUS_CODE <<- new.env(hash=TRUE)
	    lapply(names(as.list(HTTP_STATUS_CODES)),function(x){
		    code <- as.integer(x)
		    assign(x,code,CHAR_TO_STATUS_CODE)
		    assign(gsub(' ','_',tolower(HTTP_STATUS_CODES[[x]])),code,CHAR_TO_STATUS_CODE)
	    })

	    DEFAULT_SEP <<- '[&;] *'
	    callSuper(...)
	},
   escape = function(s=NULL) {
      if (is.null(s)) base::stop("Need a character vector argument")
      unlist(lapply(s,function(s){
            x <- strsplit(s,"")[[1L]]
            z <- grep('[^ a-zA-Z0-9_.-]',x,perl=TRUE)
            if (length(z)){
               y <- sapply(x[z],function(i) paste('%',paste(toupper(as.character(charToRaw(i))),collapse='%'),sep=''))
               x[z] <- y
            }
            s <- paste(x,collapse='')
            chartr(' ','+',s)
      }))
   },
   unescape = function(s=NULL){
      if(is.null(s)) base::stop("Need a character vector argument")
      unlist(lapply(s,function(s)
            utils::URLdecode(chartr('+',' ',s))
      ))
   },
   parse_query = function(qs=NULL, d=DEFAULT_SEP) {
      if (is.null(qs)) base::stop("Need a character vector argument")
      x <- strsplit(qs,d,perl=TRUE)[[1L]]
      if (length(x) == 0) return(list())

      z <- x != ''
      params <- new.env()
      params$params <- list()
      if (length(z)){
        parseEqual <- function(i,params){
          m <- regexpr('=',i)[1L]
          if (m > 0){
            if (m == 1){
              params$params[['']] <- unescape(i)
            } else { 
              paramName <- substr(i,1,m-1)
              ilen <- nchar(i)
              if (ilen == m) {
                params$params[[paramName]] <- ''
              } else {
                # handle array parameters
                paramValue <- unescape(substr(i,m+1,ilen))
                paramSet <- FALSE
                if (grepl("\\[\\]$", paramName)) {
                  paramName <- sub("\\[\\]$", "", paramName)
                  if (paramName %in% names(params$params)) {
                    params$params[[paramName]] <- c(params$params[[paramName]], paramValue)
                    paramSet <- TRUE
                  }
                }
                if (!paramSet) {
                  params$params[[paramName]] <- paramValue
                }
              }
            }
          } else {
            params$params[[i]] <- NA
          }
        }
        lapply(x[z],parseEqual,params)
      }
      params$params
   },
	#parse_nested_query = function() {},
	#normalize_params = function () {},
	build_query = function(params=NULL) {
	    if (is.null(params)) base::stop("Need named list argument")
	    # TODO: call escape here, need to vectorize it first
	    paste(names(params),params,sep='=',collapse='&')
	},
	#build_nested_query = function() {},"
	escape_html = function(string=NULL) {
	    if (is.null(string)) base::stop("Need a character vector argument")
	    string <- gsub('&','&amp;',string)
	    string <- gsub('<','&lt;',string)
	    string <- gsub('>','&gt;',string)
	    string <- gsub("'",'&#39;',string)
	    string <- gsub('"','&quot;',string)
	    string
	},
	#selectBestEncoding = function(){},
	set_cookie_header = function(header,key=NULL,value='',expires=NULL,path=NULL,domain=NULL,secure=FALSE,httpOnly=FALSE){
	    if (is.null(key)) return(invisible())
	    cookie <- paste( escape(key), paste(sapply(value,escape),collapse='&'), sep='=')
	    if (!is.null(expires))
		cookie <- paste(cookie,'; expires=',rfc2822(expires),sep='')
	    if (!is.null(path))
		cookie <- paste(cookie,'; path=',path,sep='')
	    if (!is.null(domain))
		cookie <- paste(cookie,'; domain=',domain,sep='')
	    if (secure)
		cookie <- paste(cookie,'; secure',sep='')
	    if (httpOnly)
		cookie <- paste(cookie,'; HttpOnly',sep='')

	    if (is.null(header$`Set-Cookie`) || header$`Set-Cookie` == '')
		header$`Set-Cookie` <- cookie
	    else 
		header$`Set-Cookie` <- paste(header$`Set-Cookie`,cookie,sep='\n')
	},
	delete_cookie_header = function(header,key=NULL,value='',expires=NULL,path=NULL,domain=NULL,secure=FALSE,httpOnly=FALSE){
	    if (is.null(header$`Set-Cookie`)) return(invisible())
	    cookies <- strsplit(header$`Set-Cookie`,'\n')[[1L]]
	    if (length(cookies) == 0) return(invisible())

	    d <- ifelse(is.null(domain),
		grepl(paste('^',escape(key),'=',sep=''),cookies,perl=TRUE),
		grepl(paste('^',escape(key),'=.*domain=',domain,sep=''),cookies,perl=TRUE)
	    )

	    header$`Set-Cookie` <- paste(cookies[!d],collapse='\n')

	    set_cookie_header(header,key,'',timezero(),path,domain,secure,httpOnly)
	},
	bytesize = function(string=NULL) {
	    if (is.null(string)) base::stop("Need a character vector")
	    nchar(string,type='bytes')
	},
	raw.match = function(needle=NULL,haystack=NULL,all=TRUE) .Call(Rook:::rawmatch,needle,haystack,all),
	timezero = function() structure(0,class=c('POSIXct','POSIXt')),
	rfc2822 = function(ts=NULL){
	    if (is.null(ts) || !inherits(ts,'POSIXt'))
		base::stop("Need a POSIXt object")
	    format(ts,format="%a, %d %b %Y %H:%M:%S GMT",tz='GMT')
	},
	status_code = function(status=NULL){
	    if (is.null(status) || status == '') return(as.integer(500))
	    if(is.character(status)){
		code <- try(get(gsub(' ','_',tolower(status)),CHAR_TO_STATUS_CODE,inherits=FALSE),silent=TRUE)
		ifelse(inherits(code,'try-error'),500,code)
	    } else {
		as.integer(status)
	    }
	}
    )
)$new()

#UploadedFile <- setRefClass(
#    'UploadedFile',
#)

Multipart <- setRefClass(
    'Multipart',
    fields = c('EOL'),
    methods = list(
	initialize = function(...){
	    EOL <<- '\r\n'
	    callSuper(...)
	},
	parse = function(env){
	    if (!exists('CONTENT_TYPE',env)) return(NULL)

	    # return value
	    params  <- list()

	    input <- env$`rook.input`
	    input$rewind()

	    content_length = as.integer(env$CONTENT_LENGTH)

        # Bail if not a multipart content body
        if (!grepl('multipart',env$CONTENT_TYPE))
            return(NULL)

	    # Some constants regarding boundaries
	    boundary <- paste('--',gsub('^multipart/.*boundary="?([^";,]+)"?','\\1',env$CONTENT_TYPE,perl=TRUE),sep='')
	    boundary_size <- Utils$bytesize(boundary)
	    boundaryEOL <- paste(boundary,EOL,sep='')
	    boundaryEOL_size <- boundary_size + Utils$bytesize(EOL)
	    EOLEOL = paste(EOL,EOL,sep='')
	    EOLEOL_size = Utils$bytesize(EOLEOL)
	    EOL_size = Utils$bytesize(EOL)

            buf <- new.env()
	    buf$bufsize <- 16384 # Never read more than bufsize bytes.
	    buf$read_buffer <- input$read(boundaryEOL_size)
	    buf$read_buffer_len <- length(buf$read_buffer)
	    buf$unread <- content_length - boundary_size

	    i <- Utils$raw.match(boundaryEOL,buf$read_buffer,all=FALSE)
	    if (!length(i) || i != 1){
		warning("bad content body")
		input$rewind()
		return(NULL)
	    }

	    fill_buffer <- function(x){
		buf <- input$read(ifelse(x$bufsize < x$unread, x$bufsize, x$unread))
		buflen <- length(buf)
		if (buflen > 0){
		    x$read_buffer <- c(x$read_buffer,buf)
		    x$read_buffer_len <- length(x$read_buffer)
		    x$unread <- x$unread - buflen
		}
	    }

	    # Slices off the beginning part of read_buffer.
	    slice_buffer <- function(i,size,x){
		slice <- if(i > 1) x$read_buffer[1:(i-1)] else x$read_buffer[1] 
		x$read_buffer <- if(size < x$read_buffer_len) x$read_buffer[(i+size):x$read_buffer_len] else raw()
		x$read_buffer_len <- length(x$read_buffer)
		slice
	    }

	    # prime the read_buffer
	    buf$read_buffer <- raw()
	    fill_buffer(buf)

	    while(TRUE) {
		head <- value <- NULL
		filename <- content_type <- name <- NULL
		while(is.null(head)){
		    i <- Utils$raw.match(EOLEOL,buf$read_buffer,all=FALSE)
		    if (length(i)){
			head <- slice_buffer(i,EOLEOL_size,buf)
			break
		    } else if (buf$unread){
			fill_buffer(buf)
		    } else {
			break # we've read everything and still haven't seen a valid head
		    }
		}
		if (is.null(head)){
		    warning("Bad post payload: searching for a header")
		    input$rewind()
		    return(NULL)
		} 

		# cat("Head:",rawToChar(head),"\n")
		# they're 8bit clean
		head <- rawToChar(head)

		token <- '[^\\s()<>,;:\\"\\/\\[\\]?=]+'
		condisp <- paste('Content-Disposition:\\s*',token,'\\s*',sep='')
		dispparm <- paste(';\\s*(',token,')=("(?:\\"|[^"])*"|',token,')*',sep='')

		rfc2183 <- paste('(?m)^',condisp,'(',dispparm,')+$',sep='')
		broken_quoted <- paste('(?m)^',condisp,'.*;\\sfilename="(.*?)"(?:\\s*$|\\s*;\\s*',token,'=)',sep='')
		broken_unquoted = paste('(?m)^',condisp,'.*;\\sfilename=(',token,')',sep='')

		if (length(grep(rfc2183,head,perl=TRUE))){
		    first_line <- sub(condisp,'',strsplit(head,'\r\n')[[1L]][1],perl=TRUE)
		    pairs <- strsplit(first_line,';',fixed=TRUE)[[1L]]
		    fnmatch <- '\\s*filename=(.*)\\s*'
		    if (any(grepl(fnmatch,pairs,perl=TRUE))){
			filename <- pairs[grepl(fnmatch,pairs,perl=TRUE)][1]
			filename <- gsub('"','',sub(fnmatch,'\\1',filename,perl=TRUE))
		    }
		} else if (length(grep(broken_quoted,head,perl=TRUE))){
		    filename <- sub(broken_quoted,'\\1',strsplit(head,'\r\n')[[1L]][1],perl=TRUE)
		} else if (length(grep(broken_unquoted,head,perl=TRUE))){
		    filename <- sub(broken_unquoted,'\\1',strsplit(head,'\r\n')[[1L]][1],perl=TRUE)
		}

		if (!is.null(filename) && filename!=''){
		    filename = Utils$unescape(filename)
		}

		headlines <- strsplit(head,EOL,fixed=TRUE)[[1L]]
		content_type_re <- '(?mi)Content-Type: (.*)'
		content_types <- headlines[grepl(content_type_re,headlines,perl=TRUE)]
		if (length(content_types)){
		    content_type <- sub(content_type_re,'\\1',content_types[1],perl=TRUE)
		}

		name <- sub('(?si)Content-Disposition:.*\\s+name="?([^";]*).*"?','\\1',head,perl=TRUE)

		while(TRUE){
		    i <- Utils$raw.match(boundary,buf$read_buffer,all=FALSE)
		    if (length(i)){
			value <- slice_buffer(i,boundary_size,buf)
			if (length(value)){

			    # Drop EOL only values
			    if (length(value) == 2 && length(Utils$raw.match(EOL,value)))
				break

			    if (!is.null(filename) || !is.null(content_type)){
				data <- list()
				if (!is.null(filename))
				    data$filename <- strsplit(filename,'[\\/]',perl=TRUE)[[1L]]
				data$tempfile <- tempfile('Multipart')
				if (!is.null(content_type))
				    data$content_type <- content_type
				data$head <- head
				con <- file(data$tempfile,open='wb')
				writeBin(value,con)
				close(con)
				params[[name]] <- data
			    } else {
				len <- length(value)
				# Trim trailing EOL
				if (len > 2 && length(Utils$raw.match(EOL,value[(len-1):len],all=FALSE)))
				    len <- len -2

				# handle array parameters
				paramValue <- Utils$escape(rawToChar(value[1:len]))
				paramSet <- FALSE
				if (grepl("\\[\\]$", name)) {
				   name <- sub("\\[\\]$", "", name)
				   if (name %in% names(params)) {
					params[[name]] <- c(params[[name]], paramValue)
					paramSet <- TRUE
				   }
				}
				if (!paramSet) {
				   params[[name]] <- paramValue
				}
			    }
			} 
			break
		    } else if (buf$unread){
			fill_buffer(buf)
		    } else {
			break # we've read everything and still haven't seen a valid value
		    }
		}
		if (is.null(value)){
		    # bad post payload
		    input$rewind()
		    warning("Bad post payload: searching for a body part")
		    return(NULL)
		}

		# Now search for ending markers or the beginning of another part
		while (buf$read_buffer_len < 2 && buf$unread) fill_buffer(buf)

		if (buf$read_buffer_len < 2 && buf$unread == 0){
		    # Bad stuff at the end. just return what we've got
		    # and presume everything is okay.
		    input$rewind()
		    return(params)
		}

		# Valid ending
		if (length(Utils$raw.match('--',buf$read_buffer[1:2],all=FALSE))){
		    input$rewind()
		    return(params)
		} 
		# Skip past the EOL.
		if (length(Utils$raw.match(EOL,buf$read_buffer[1:EOL_size],all=FALSE))){
		    slice_buffer(1,EOL_size,buf)
		} else {
		    warning("Bad post body: EOL not present")
		    input$rewind()
		    return(params)
		}

		# another sanity check before we try to parse another part
		if ((buf$read_buffer_len + buf$unread) < boundary_size){
		    warning("Bad post body: unknown trailing bytes")
		    input$rewind()
		    return(params)
		}
	    }
	},
	build = function(params){
	}
    )
)$new()

Mime <- setRefClass(
    'Mime',
    fields = 'MIME_TYPES',
    methods = list(
	initialize = function(...){
	    MIME_TYPES <<- new.env(hash=TRUE)
	    with(MIME_TYPES,{
	      ".3gp"     <- "video/3gpp"
	      ".a"       <- "application/octet-stream"
	      ".ai"      <- "application/postscript"
	      ".aif"     <- "audio/x-aiff"
	      ".aiff"    <- "audio/x-aiff"
	      ".asc"     <- "application/pgp-signature"
	      ".asf"     <- "video/x-ms-asf"
	      ".asm"     <- "text/x-asm"
	      ".asx"     <- "video/x-ms-asf"
	      ".atom"    <- "application/atom+xml"
	      ".au"      <- "audio/basic"
	      ".avi"     <- "video/x-msvideo"
	      ".bat"     <- "application/x-msdownload"
	      ".bin"     <- "application/octet-stream"
	      ".bmp"     <- "image/bmp"
	      ".bz2"     <- "application/x-bzip2"
	      ".c"       <- "text/x-c"
	      ".cab"     <- "application/vnd.ms-cab-compressed"
	      ".cc"      <- "text/x-c"
	      ".chm"     <- "application/vnd.ms-htmlhelp"
	      ".class"   <- "application/octet-stream"
	      ".com"     <- "application/x-msdownload"
	      ".conf"    <- "text/plain"
	      ".cpp"     <- "text/x-c"
	      ".crt"     <- "application/x-x509-ca-cert"
	      ".css"     <- "text/css"
	      ".csv"     <- "text/csv"
	      ".cxx"     <- "text/x-c"
	      ".deb"     <- "application/x-debian-package"
	      ".der"     <- "application/x-x509-ca-cert"
	      ".diff"    <- "text/x-diff"
	      ".djv"     <- "image/vnd.djvu"
	      ".djvu"    <- "image/vnd.djvu"
	      ".dll"     <- "application/x-msdownload"
	      ".dmg"     <- "application/octet-stream"
	      ".doc"     <- "application/msword"
	      ".dot"     <- "application/msword"
	      ".dtd"     <- "application/xml-dtd"
	      ".dvi"     <- "application/x-dvi"
	      ".ear"     <- "application/java-archive"
	      ".eml"     <- "message/rfc822"
	      ".eps"     <- "application/postscript"
	      ".exe"     <- "application/x-msdownload"
	      ".f"       <- "text/x-fortran"
	      ".f77"     <- "text/x-fortran"
	      ".f90"     <- "text/x-fortran"
	      ".flv"     <- "video/x-flv"
	      ".for"     <- "text/x-fortran"
	      ".gem"     <- "application/octet-stream"
	      ".gemspec" <- "text/x-script.ruby"
	      ".gif"     <- "image/gif"
	      ".gz"      <- "application/x-gzip"
	      ".h"       <- "text/x-c"
	      ".htc"     <- "text/x-component"
	      ".hh"      <- "text/x-c"
	      ".htm"     <- "text/html"
	      ".html"    <- "text/html"
	      ".ico"     <- "image/vnd.microsoft.icon"
	      ".ics"     <- "text/calendar"
	      ".ifb"     <- "text/calendar"
	      ".iso"     <- "application/octet-stream"
	      ".jar"     <- "application/java-archive"
	      ".java"    <- "text/x-java-source"
	      ".jnlp"    <- "application/x-java-jnlp-file"
	      ".jpeg"    <- "image/jpeg"
	      ".jpg"     <- "image/jpeg"
	      ".js"      <- "application/javascript"
	      ".json"    <- "application/json"
	      ".log"     <- "text/plain"
	      ".m3u"     <- "audio/x-mpegurl"
	      ".m4v"     <- "video/mp4"
	      ".man"     <- "text/troff"
	      ".manifest"<- "text/cache-manifest"
	      ".mathml"  <- "application/mathml+xml"
	      ".mbox"    <- "application/mbox"
	      ".mdoc"    <- "text/troff"
	      ".me"      <- "text/troff"
	      ".mid"     <- "audio/midi"
	      ".midi"    <- "audio/midi"
	      ".mime"    <- "message/rfc822"
	      ".mml"     <- "application/mathml+xml"
	      ".mng"     <- "video/x-mng"
	      ".mov"     <- "video/quicktime"
	      ".mp3"     <- "audio/mpeg"
	      ".mp4"     <- "video/mp4"
	      ".mp4v"    <- "video/mp4"
	      ".mpeg"    <- "video/mpeg"
	      ".mpg"     <- "video/mpeg"
	      ".ms"      <- "text/troff"
	      ".msi"     <- "application/x-msdownload"
	      ".odp"     <- "application/vnd.oasis.opendocument.presentation"
	      ".ods"     <- "application/vnd.oasis.opendocument.spreadsheet"
	      ".odt"     <- "application/vnd.oasis.opendocument.text"
	      ".ogg"     <- "application/ogg"
	      ".ogv"     <- "video/ogg"
	      ".p"       <- "text/x-pascal"
	      ".pas"     <- "text/x-pascal"
	      ".pbm"     <- "image/x-portable-bitmap"
	      ".pdf"     <- "application/pdf"
	      ".pem"     <- "application/x-x509-ca-cert"
	      ".pgm"     <- "image/x-portable-graymap"
	      ".pgp"     <- "application/pgp-encrypted"
	      ".pkg"     <- "application/octet-stream"
	      ".pl"      <- "text/x-script.perl"
	      ".pm"      <- "text/x-script.perl-module"
	      ".png"     <- "image/png"
	      ".pnm"     <- "image/x-portable-anymap"
	      ".ppm"     <- "image/x-portable-pixmap"
	      ".pps"     <- "application/vnd.ms-powerpoint"
	      ".ppt"     <- "application/vnd.ms-powerpoint"
	      ".ps"      <- "application/postscript"
	      ".psd"     <- "image/vnd.adobe.photoshop"
	      ".py"      <- "text/x-script.python"
	      ".qt"      <- "video/quicktime"
	      ".ra"      <- "audio/x-pn-realaudio"
	      ".rake"    <- "text/x-script.ruby"
	      ".ram"     <- "audio/x-pn-realaudio"
	      ".rar"     <- "application/x-rar-compressed"
	      ".rb"      <- "text/x-script.ruby"
	      ".rdf"     <- "application/rdf+xml"
	      ".roff"    <- "text/troff"
	      ".rpm"     <- "application/x-redhat-package-manager"
	      ".rss"     <- "application/rss+xml"
	      ".rtf"     <- "application/rtf"
	      ".ru"      <- "text/x-script.ruby"
	      ".s"       <- "text/x-asm"
	      ".sgm"     <- "text/sgml"
	      ".sgml"    <- "text/sgml"
	      ".sh"      <- "application/x-sh"
	      ".sig"     <- "application/pgp-signature"
	      ".snd"     <- "audio/basic"
	      ".so"      <- "application/octet-stream"
	      ".svg"     <- "image/svg+xml"
	      ".svgz"    <- "image/svg+xml"
	      ".swf"     <- "application/x-shockwave-flash"
	      ".t"       <- "text/troff"
	      ".tar"     <- "application/x-tar"
	      ".tbz"     <- "application/x-bzip-compressed-tar"
	      ".tcl"     <- "application/x-tcl"
	      ".tex"     <- "application/x-tex"
	      ".texi"    <- "application/x-texinfo"
	      ".texinfo" <- "application/x-texinfo"
	      ".text"    <- "text/plain"
	      ".tif"     <- "image/tiff"
	      ".tiff"    <- "image/tiff"
	      ".torrent" <- "application/x-bittorrent"
	      ".tr"      <- "text/troff"
	      ".txt"     <- "text/plain"
	      ".vcf"     <- "text/x-vcard"
	      ".vcs"     <- "text/x-vcalendar"
	      ".vrml"    <- "model/vrml"
	      ".war"     <- "application/java-archive"
	      ".wav"     <- "audio/x-wav"
	      ".webm"    <- "video/webm"
	      ".wma"     <- "audio/x-ms-wma"
	      ".wmv"     <- "video/x-ms-wmv"
	      ".wmx"     <- "video/x-ms-wmx"
	      ".wrl"     <- "model/vrml"
	      ".wsdl"    <- "application/wsdl+xml"
	      ".xbm"     <- "image/x-xbitmap"
	      ".xhtml"   <- "application/xhtml+xml"
	      ".xls"     <- "application/vnd.ms-excel"
	      ".xml"     <- "application/xml"
	      ".xpm"     <- "image/x-xpixmap"
	      ".xsl"     <- "application/xml"
	      ".xslt"    <- "application/xslt+xml"
	      ".yaml"    <- "text/yaml"
	      ".yml"     <- "text/yaml"
	      ".zip"     <- "application/zip"
	    })
	    callSuper(...)
	},
	file_extname = function(fname=NULL){
	    if (is.null(fname))
		base::stop("need an argument of character")
	    paste('.',rev(strsplit(fname,'.',fixed=TRUE)[[1]])[1],sep='')
	},
	mime_type=function(ext=NULL,fallback='application/octet-stream'){
	    if (is.null(ext) || !nzchar(ext) || is.null(MIME_TYPES[[ext]]))
		fallback
	    else
		MIME_TYPES[[ext]]
	}
    )
)$new()

is_rookable <- function(app){
    if (is(app,'refClass')) TRUE
    else if (is(app,'function')) TRUE
    else FALSE
}

# Abstract Rook::App and Rook::Middleware that Builder and related apps inherit from.
App <- setRefClass(
    'App',
    fields = c('app'),
    methods = list(
	initialize = function(app=NULL,...) {
	    app <<- app
	    callSuper(...)
	},
	call = function(env){
	    if (is(app,'refClass')) app$call(env)
	    else if (is(app,'function')) app(env)
	    else stop('App not Rook aware')
	}
    )
)

Middleware <- setRefClass(
    'Middleware',
    contains = 'App',
    methods = list(
	initialize = function(...) {
	    callSuper(...)
	},
	set_app = function(app){
	    app <<- app
	}
    )
)

Builder <- setRefClass(
    'Builder',
    contains = 'App',
    methods = list(
	initialize = function(...){
	    objs <- list(...)
	    if (length(objs) > 1){
		for (i in 1:(length(objs)-1)){
		    if (inherits(objs[[i]],'Middleware'))
			objs[[i]]$set_app(objs[[i+1]])
		    else
			stop("Argument ",i,"is not a Middleware object")
		}
	    }
	    callSuper(app=objs[[1]])
	}
    )
)

Brewery <- setRefClass(
   'Brewery',
   contains = 'Middleware',
   fields = c('url','root','opt'),
   methods = list(
      initialize = function(url,root,...){
         library("brew")
         url <<- sub("/+$","",url)
         root <<- normalizePath(root,mustWork=TRUE)
         opts <- list(...)
         if (length(opts)>0){
            opt <<- try(list2env(opts),silent=TRUE)
            if (inherits(opt,'try-error'))
               stop('Optional arguments must be named')
         } else {
            opt <<- new.env()
         }
         callSuper()
      },
      call = function(env){
         req <- Rook::Request$new(env)
         res <- Rook::Response$new()
         opt[['req']] <<- req
         opt[['res']] <<- res
         path <- env[["PATH_INFO"]]
         file_path <- try(
            normalizePath(file.path(root,path),mustWork=TRUE),
            silent=TRUE
         )
         if (!inherits(file_path, 'try-error') &&
             grepl(paste('^',url,sep=''),path) &&
             !grepl(paste('^',url,'$',sep=''),path) &&
             grepl(paste('^',root,url,.Platform$file.sep,sep=''),file_path)){

            oldwd <- setwd(dirname(file_path))
            on.exit(setwd(oldwd))
            res$write(
               paste(capture.output(brew::brew(basename(file_path),envir=opt)),
                  collapse="\n")
               )
            res$finish()
         } else {
            app$call(env)
         }
      }
   )
)

Request <- setRefClass(
    'Request',
    fields = c('FORM_DATA_MEDIA_TYPES','PARSEABLE_DATA_MEDIA_TYPES','env'),
    methods = list(
	initialize = function(env,...){
	    env <<- env

	    # The set of form-data media-types. Requests that do not indicate
	    # one of the media types presents in this list will not be eligible
	    # for form-data / param parsing.
	    FORM_DATA_MEDIA_TYPES <<- c(
	      'application/x-www-form-urlencoded',
	      'multipart/form-data'
	    )

	    # The set of media-types. Requests that do not indicate
	    # one of the media types presents in this list will not be eligible
	    # for param parsing like soap attachments or generic multiparts.
	    PARSEABLE_DATA_MEDIA_TYPES <<- c(
	      'multipart/related',
	      'multipart/mixed'
	    )

	    # Adjust script name if we're sitting behind a reverse proxy
	    #
	    # Nginx example config:
	    #
	    #	location /rhttpd {
	    #	    rewrite ^/rhttpd(.*)$  $1 break;
	    #	    proxy_pass http://127.0.0.1:12344;
	    #	    proxy_set_header X-Script-Name /rhttpd;
	    #	}
	    #
	    
	    if (exists('HTTP_X_SCRIPT_NAME',env)){
		env[['HTTP_X_SCRIPT_NAME']] <<- sub('/$','',env[['HTTP_X_SCRIPT_NAME']])
		env[['SCRIPT_NAME']] <<- paste(env[['HTTP_X_SCRIPT_NAME']],env[['SCRIPT_NAME']],sep='')
	    } 

	    callSuper(...)
	},
	body = function()            env[["rook.input"]],
	scheme = function()          env[["rook.url_scheme"]],
	path_info = function()       env[["PATH_INFO"]],
	port = function()            as.integer(env[["SERVER_PORT"]]),
	request_method = function()  env[["REQUEST_METHOD"]],
	query_string = function()    env[["QUERY_STRING"]],
	content_length = function()  env[['CONTENT_LENGTH']],
	content_type = function()    env[['CONTENT_TYPE']],
	media_type = function(){
	    if (is.null(content_type())) return(NULL)
	    tolower(strsplit(content_type(),'\\s*[;,]\\s*')[[1]][1])
	},
	media_type_params = function(){
	    if (is.null(content_type())) return(NULL)
	    params <- list()
	    for(i in strsplit(content_type(),'\\s*[;,]\\s*')[[1]][-1]){
		x <- strsplit(i,'=')[[1]]
		params[[tolower(x[1])]] <- x[2]
	    }
	    params
	},
	content_charset = function() media_type_params()[['charset']],
	host_with_port = function(){
	    if(exists('HTTP_X_FORWARDED_HOST',env)){
		x <- strsplit(env[['HTTP_X_FORWARDED_HOST']],',\\s?')[[1]]
		return(x[length(x)])
	    } else if (exists('HTTP_HOST',env)){
		env[['HTTP_HOST']]
	    } else {
		if (exists('SERVER_NAME',env))
		    host <- env[['SERVER_NAME']]
		else
		    host <- env[['SERVER_ADDR']]
		paste(host,env[['SERVER_PORT']],sep=':')
	    }
	},
	host = function() sub(':\\d+','',host_with_port(),perl=TRUE),
	script_name = function(s=NULL){
	    if (!is.null(s) && is.character(s)) env[['SCRIPT_NAME']] <<- s
	    env[['SCRIPT_NAME']]
	},
	path_info = function(s=NULL){
	    if (!is.null(s) && is.character(s)) env[['PATH_INFO']] <<- s
	    env[['PATH_INFO']]
	},
	delete = function()	request_method() == 'DELETE',
	get = function()	request_method() == 'GET',
	head = function()	request_method() == 'HEAD',
	options = function()	request_method() == 'OPTIONS',
	post = function()	request_method() == 'POST',
	put = function()	request_method() == 'PUT',
	trace = function()	request_method() == 'TRACE',
	form_data = function(){
	    (post() && !is.null(media_type())) || any(FORM_DATA_MEDIA_TYPES==media_type())
	},
	parseable_data = function(){
	    any(PARSEABLE_DATA_MEDIA_TYPES==media_type())
	},
	GET = function(){
	    if (!exists('rook.request.query_list',env))
		env[['rook.request.query_list']] <<- Utils$parse_query(query_string())
	    env[['rook.request.query_list']]
	},
	POST = function(){
	    if (!exists('rook.input',env))
		stop("Missing rook.input")
	    if (exists('rook.request.form_list',env))
		env[['rook.request.form_list']]
	    else if (form_data() || parseable_data()){
		env[['rook.request.form_list']] <<- Multipart$parse(env)
		if (length(env[['rook.request.form_list']]) == 0){
		    form_vars <- env[['rook.input']]$read()
		    env[['rook.request.form_list']] <<- Utils$parse_query(rawToChar(form_vars))
		}
	    }
	    env[['rook.request.form_list']]
	},
	params = function() c(GET(),POST()) ,
	referer = function(){
	    if (!is.null(env[['HTTP_REFERER']])) env[['HTTP_REFERER']] else '/'
	},
	referrer = function() referer(),
	user_agent = function() env[['HTTP_USER_AGENT']],
	cookies = function(){
	    if (exists('rook.request.cookie_list',env))
		return(env[['rook.request.cookie_list']])

	    if (!is.null(env[['HTTP_COOKIE']]))
		env[['rook.request.cookie_list']] <<- Utils$parse_query(env[['HTTP_COOKIE']])
	    else
		env[['rook.request.cookie_list']] <<- NULL
	},
	xhr = function() {
	    (exists('HTTP_X_REQUESTED_WITH',env) && 
	    env[['HTTP_X_REQUESTED_WITH']] == 'XMLHttpRequest')
	},
	url = function(){
	    x <- paste(scheme(),'://',host(),sep='')
	    if ( (scheme() == 'https' && port() != 443) || (scheme() == 'http' && port() != 80))
		x <- paste(x,':',port(),sep='')
	    x <- paste(x,fullpath(),sep='')
	    x
	},
	path = function() paste(script_name(),path_info(),sep=''),
	fullpath = function(){
	    if (is.null(query_string()))
		path()
	    else
		paste(path(),'?',query_string(),sep='')
	},
	to_url = function(url,...) {
	    newurl <- paste(script_name(),url,sep='')
	    opt <- list(...)
	    if (length(opt)){
		    newurl <- paste(
			newurl,'?',
			paste(names(opt),opt,sep='=',collapse='&'),
			sep=''
		    )
	    }
	    newurl
	},
	accept_encoding = function() env[['HTTP_ACCEPT_ENCODING']],
	ip = function() env[['REMOTE_ADDR']]
    )
)

suspend_console <- function(){

   # My first approach to suspending the console
   # was to call Sys.sleep with .Machine$integer.max, however
   # this fails on windows because its implementating causes
   # overflow.

   # So we sleep for a day, and then loop.
   allday <- 24*60*60
   while(TRUE) Sys.sleep(allday)
}

URLMap <- setRefClass(
    'URLMap',
    fields = c('map'),
    methods = list(
	initialize = function(...){
	    apps <- list(...)
	    map <<- list()
	    for (url in names(apps)){
		x <- apps[[url]]
		if (is(x,'function'))
		    map[[length(map)+1]] <<- App$new(x)
		else if (is_rookable(x))
		    map[[length(map)+1]] <<- x
		else
		    stop("App for url ",url," not rookable")
	    }
	    names(map) <<- names(apps)
	    callSuper()
	},
	call = function(env){
	    path <- env[['PATH_INFO']]
	    for (url in names(map)){
	        #cat('matching url',url,'to',path,'\n')
		if (grepl(url,path))
		    return(map[[url]]$call(env))
	    }
	    body <- paste("File not found:",path,"\n")
	    list(
		status=404L, 
		headers = list(
		    "Content-Type" = "text/plain",
		    "Content-Length" = as.character(nchar(body)),
		    "X-Cascade" = "pass"
		),
		body = body
	    )
	}
    )
)

\name{suspend_console}
\alias{suspend_console}
\title{
Suspend the R console 
}
\description{
Calls Sys.sleep in a never-ending while loop to mimic suspension of the R console.
}
\usage{
suspend_console()
}
\value{
    No value is ever returned.
}
\seealso{
\code{\link{Rook}}.
}
\keyword{function}

\name{Middleware-class}
\Rdversion{1.1}
\docType{class}
\alias{Middleware-class}
\alias{Middleware}

\title{Class \code{Middleware}}
\description{
An abstract class for building Rook Middleware applications. \code{Middleware} applications either handle the incoming web request or hand off the request to the Rook app defined in the field of the same name.
}
\examples{
# Middleware applications are typically instantiated in the argument list of
# Builder$new(), but here is stand-alone example.
# 
# Once your browser loads the app, you will see something like this in
# your location bar: http://127.0.0.1:28649/custom/middle. Add '/foo'
# onto the end of that and reload.

setRefClass(
    'FooBar',
    contains = 'Middleware',
    methods = list(
        initialize = function(...){
            # app to defer to.
            callSuper(app=App$new(function(env){
                res <- Response$new()
                res$write("<h1>I'm the deferred app.</h1>")
                res$finish()
            }))
        },
        call = function(env){
            req <- Request$new(env)
            res <- Response$new()
            if (length(grep('foo',req$path_info()))){
                res$write("<h1>I'm the middleware app.</h1>")        
                return(res$finish())
            } else {
                app$call(env)
            }
        }
    )
)
s <- Rhttpd$new()
\dontrun{
s$start(quiet=TRUE)
}
s$add(name="middle",app=getRefClass('FooBar')$new())
\dontrun{
s$browse('middle') # Opens a browser window to the app.
}
s$remove(all=TRUE)
rm(s)
}
\seealso{
The following classes implement Middleware: 
    \code{\link{Brewery}} and \code{\link{Static}}.
}
\section{Methods}{
  \describe{
    \item{\code{set_app(app)}:}{ \code{app} is a \code{\link{Rook}} application that will handle the request if this Middleware app does not. }
  }
}
\keyword{classes}

\name{RhttpdApp-class}
\Rdversion{1.1}
\docType{class}
\alias{RhttpdApp-class}
\alias{RhttpdApp}

\title{Class \code{RhttpdApp}}
\description{
Creates a Rook application ready to add to an \code{\link{Rhttpd}} server.
}
\details{

The internal web server allows dispatching to user-defined closures
located in tools:::.httpd.handlers.env. For instance, if a handler named
'foo' is placed there, then the url path to that handler is /custom/foo.

\code{RhttpdApp} along with \code{\link{Rhttpd}} hide these details by
allowing a user to create application objects specifying only their name
and the application. There is currently a limit of 63 characters
or less for application names.

NOTE: When a file is given as the value of the \code{app} argument
to \code{new()}, it is monitored for timestamp changes. If a change
occurs in the modification time as returned by \code{\link[base]{file.info}},
then the file is sourced prior to handling subsequent requests.

}
\seealso{
\code{\link{Rhttpd}}.
}
\examples{

s <- Rhttpd$new()
s$add(RhttpdApp$new(
    name='summary',
    app=system.file('exampleApps/summary.R',package='Rook')
))
\dontrun{
s$start(quiet=TRUE)
s$browse(1)
}
s$remove(all=TRUE)

#  Stops the server but doesn't uninstall the app
\dontrun{
s$stop()
}
s$remove(all=TRUE)
rm(s)
}
\keyword{classes}
\section{Methods}{
  \describe{
    \item{\code{new(app, name)}:}{ Creates an object of class \code{RhttpdApp}. Argument \code{app} can be any \code{\link{Rook}} aware object or it can be a location to a file whose source creates a Rook aware object. That object must be named either \code{'app'} or the value of \code{name}. \code{name} is a character vector.}
  }
}

\name{Server}
\Rdversion{1.1}
\alias{Server}

\title{Rook Server Object}
\description{
\code{Server} is an object exported by Rook that has no value to the user. It is mainly used by web servers for their convenience. To see an example of how it may be used, see rApache.R in the inst/servers directory.
} 
\keyword{classes}

\name{Request-class}
\Rdversion{1.1}
\docType{class}
\alias{Request-class}
\alias{Request}

\title{Class \code{Request}}
\description{
A convenience class for working with a \code{\link{Rook}} environment. Be sure to see the example at the end of this help file.
}
\examples{
#
# The following example prints out the result of each method.
# 
ls_str <- function(s) paste(capture.output(str(s),file=NULL),collapse='\n')
s <- Rhttpd$new()
\dontrun{
s$start(quiet=TRUE)
}
s$add(name="request",
    app=function(env){
        req <- Request$new(env)
        res <- Response$new()
        res$set_cookie('imacookie','42')
        action <- req$to_url('/foo',bar=1,baz='three')
        res$write('<form enctype="multipart/form-data" method=POST action="')
        res$write(action)
        res$write('">')
        res$write('Upload a file: <input type=file name=fileUpload>')
        res$write('<input type=submit></form><br><pre>')
        res$write(c('parseable_data: ',req$parseable_data(),'\n'))
        res$write(c('url: ',req$url(),'\n'))
        res$write(c('request_method: ',req$request_method(),'\n'))
        res$write(c('GET: ',ls_str(req$GET()),'\n'))
        res$write(c('post: ',req$post(),'\n'))
        res$write(c('media_type: ',req$media_type(),'\n'))
        res$write(c('query_string: ',req$query_string(),'\n'))
        res$write(c('fullpath: ',req$fullpath(),'\n'))
        res$write(c('referer: ',req$referer(),'\n'))
        res$write(c('cookies: ',ls_str(req$cookies()),'\n'))
        res$write(c('content_charset: ',req$content_charset(),'\n'))
        res$write(c('head: ',req$head(),'\n'))
        res$write(c('accept_encoding: ',req$accept_encoding(),'\n'))
        res$write(c('content_length: ',req$content_length(),'\n'))
        res$write(c('form_data: ',req$form_data(),'\n'))
        res$write(c('xhr: ',req$xhr(),'\n'))
        res$write(c('params: ',ls_str(req$params()),'\n'))
        res$write(c('media_type_params:\n',ls_str(req$media_type_params()),'\n'))
        res$write(c('user_agent: ',req$user_agent(),'\n'))
        res$write(c('put: ',req$put(),'\n'))
        res$write(c('get: ',req$get(),'\n'))
        res$write(c('path: ',req$path(),'\n'))
        res$write(c('body: ',ls_str(req$body()),'\n'))
        res$write(c('port: ',req$port(),'\n'))
        res$write(c('host_with_port: ',req$host_with_port(),'\n'))
        res$write(c('scheme: ',req$scheme(),'\n'))
        res$write(c('ip: ',req$ip(),'\n'))
        res$write(c('options: ',req$options(),'\n'))
        res$write(c('to_url: ',req$to_url('foo',bar=1,baz='two'),'\n'))
        res$write(c('host: ',req$host(),'\n'))
        res$write(c('POST: ',ls_str(req$POST()),'\n'))
        res$write(c('trace: ',req$trace(),'\n'))
        res$write(c('script_name: ',req$script_name(),'\n'))
        res$write(c('content_type: ',req$content_type(),'\n'))
        res$write(c('delete: ',req$delete(),'\n'))
        res$write(c('path_info: ',req$path_info(),'\n'))
        res$write(c('\nRac env: ',ls_str(as.list(env)),'\n'))
        res$finish()
    }
)
\dontrun{
s$browse('request') # Opens a browser window to the app.
}
s$remove(all=TRUE)
rm(s)
}
\section{Methods}{
  \describe{
    \item{\code{parseable_data()}:}{ Returns a boolean value determining if the POST payload is parseable. }
    \item{\code{url()}:}{ Returns url as a character string containing the scheme, host, port, and possibly the GET query string if supplied.}
    \item{\code{request_method()}:}{ Returns the HTTP method as a character string, e.g. 'GET', 'POST', etc. }
    \item{\code{GET()}:}{ Returns a named list containing the variables parsed from the query string. }
    \item{\code{post()}:}{ Returns TRUE if the current request method is 'POST', FALSE otherwise. }
    \item{\code{new(env)}:}{ Instantiates a new \code{Request} object for the given \code{Rook} environment.}
    \item{\code{media_type()}:}{ Returns the media type for the current request as a character string.}
    \item{\code{query_string()}:}{ Returns the unparsed query string. }
    \item{\code{fullpath()}:}{ Returns the same string as url() but without the scheme, host, and port. }
    \item{\code{referer()} or \code{referrer()}:}{ Returns the referring url.  }
    \item{\code{cookies()}:}{ Returns any cookies in the request as a named list. }
    \item{\code{content_charset()}:}{ Returns the content charset as a character string. }
    \item{\code{head()}:}{ Returns TRUE if the HTTP method is 'HEAD', FALSE otherwise.}
    \item{\code{accept_encoding()}:}{ Returns the accept encoding header as a character string.}
    \item{\code{content_length()}:}{ Returns content length header value as a string. }
    \item{\code{form_data()}:}{ Returns TRUE if there's form data, e.g. POST data with the request, FALSE otherwise.}
    \item{\code{xhr()}:}{ Returns the x-requested-with header value as a character string.}
    \item{\code{params()}:}{ Returns the combination of \code{POST()} and \code{GET()} in one named list.}
    \item{\code{media_type_params()}:}{ Returns any media type parameters from the content type as a named list. }
    \item{\code{user_agent()}:}{ Returns the user-agent header value as a character string. }
    \item{\code{put()}:}{ Returns TRUE if the current request is a 'PUT'. }
    \item{\code{get()}:}{ Returns TRUE if the current request is a 'GET'. }
    \item{\code{path()}:}{ Returns a character string like \code{fullpath()} but without the query string. }
    \item{\code{body()}:}{ Returns the 'rook.input' object from the environment. See \code{\link{RhttpdInputStream}} for more information.}
    \item{\code{port()}:}{ Returns the server port as an integer.e}
    \item{\code{host_with_port()}:}{ Returns the host and port as a character string separated by ':'. }
    \item{\code{scheme()}:}{ Returns the scheme, e.g. 'http' or 'https', as a character string. }
    \item{\code{ip()}:}{ Returns the remote IP address as a character string. }
    \item{\code{options()}:}{ Returns TRUE if the current request is 'OPTIONS'. }
    \item{\code{to_url(url, ...)}:}{ Concatenates the script name with the \code{url} argument along with any named parameters passed via \code{...} .}
    \item{\code{host()}:}{ Returns the server host as a character string. }
    \item{\code{POST()}:}{ Returns a named list containing the variables parsed from the POST payload.}
    \item{\code{trace()}:}{ Returns TRUE if the current request is 'TRACE'. }
    \item{\code{script_name(s=NULL)}:}{ Returns the script name of the application, e.g. '/custom/multi'. Also, if \code{s} is not NULL, sets the script name to \code{s}. }
    \item{\code{content_type()}:}{ Returns the content-type header value as a character string. }
    \item{\code{delete()}:}{ Returns TRUE if the current request is 'DELETE'. }
    \item{\code{path_info(s=NULL)}:}{ Returns the portion of the url after the script name as a character string. If \code{s} is not NULL, sets the path info to \code{s}.}
  }
}
\seealso{
\code{\link{Rhttpd}} and \code{\link{Response}}.
}
\keyword{classes}

\name{Mime-class}
\Rdversion{1.1}
\docType{class}
\alias{Mime-class}
\alias{Mime}

\title{Class \code{Mime} and object \code{Mime}}
\description{
A convenience object for determining the MIME type of a file name.
}
\examples{
Mime$file_extname('foo.png')
Mime$mime_type('.png')
}
\keyword{classes}
\section{Methods}{
  \describe{
    \item{\code{file_extname(fname=NULL)}:}{ Returns the file extensions for the given file.}
    \item{\code{mime_type(ext=NULL, fallback='application/octet-stream')}:}{ Returns the MIME type given the file extension. Be sure to include the dot character in \code{ext}. If no match is found, then the fallback MIME type is returned.}
  }
}

\name{File-class}
\Rdversion{1.1}
\docType{class}
\alias{File-class}
\alias{File}

\title{Class \code{File}}
\description{
A Rook application that serves static files from a root directory, according to the path info of the Rook request.
}
\examples{
# This example serves all your files in /etc (on UNIX and Mac only).
#
# Note that when you open the application, you will see the word
# 'Forbidden'. "File" doesn't serve directories, so you must amend the
# url in the location bar with the file you want to view. Try adding /passwd.

s <- Rhttpd$new()
\dontrun{
s$start(quiet=TRUE)
}
s$add(name="etc",app=File$new('/etc'))
\dontrun{
s$browse('etc') # Opens a browser window to the app.
}
s$remove(all=TRUE)
rm(s)
}
\section{Methods}{
\describe{
    \item{\code{new(root):}}{
	\code{root} is the name of the directory from where to serve files.
    }
}}
\seealso{
\code{\link{Rhttpd}}.
}
\keyword{classes}

\name{Multipart-class}
\Rdversion{1.1}
\docType{class}
\alias{Multipart-class}
\alias{Multipart}

\title{Class \code{Multipart} and object \code{Multipart}}
\description{
A convenience object for parsing multipart/form-data POST payloads.
}
\examples{
s <- Rhttpd$new()
\dontrun{
s$start(quiet=TRUE)
}
s$add(name="multi",
    app=function(env){
        req <- Request$new(env)
        res <- Response$new()
        res$write('<form enctype="multipart/form-data" method=POST>')
        res$write('Upload a file: <input type=file name=fileUpload>')
        res$write('<input type=submit></form><br>')
        post <- Multipart$parse(env)
        if (length(post)){
            poststr <- paste(capture.output(str(post),file=NULL),collapse='\n')
            res$write(c('<pre>',poststr,'</pre>'))
        }
        res$finish()
    }
)
\dontrun{
s$browse('multi') # Opens a browser window to the app.
}
s$remove(all=TRUE)
rm(s)
}
\seealso{
\code{\link{Rhttpd}}, \code{\link{Request}}, and \code{\link{Response}}.
}
\section{Methods}{
  \describe{
    \item{\code{parse(env)}:}{ Returns parsed POST payload as a named list. \code{env} is an environment created by \code{Rhttpd} and conforms to the \code{\link{Rook}} specification.}
  }
}
\keyword{classes}

\name{Response-class}
\Rdversion{1.1}
\docType{class}
\alias{Response-class}
\alias{Response}

\title{Class \code{Response}}
\description{
A convenience class for creating \code{\link{Rook}} responses.
}
\seealso{
\code{\link{Rhttpd}} and \code{\link{Request}}.
}
\examples{
s <- Rhttpd$new()
\dontrun{
s$start(quiet=TRUE)
}
s$add(name="response",
    app=function(env){
        req <- Request$new(env)
        res <- Response$new()
        res$write('hello')
        res$finish()
    }
)
\dontrun{
s$browse('response') # Opens a browser window to the app.
}
s$remove(all=TRUE)
rm(s)
}
\keyword{classes}
\section{Methods}{
  \describe{
    \item{\code{header(key, value)}:}{ Sets an HTTP header for the
      response. Both \code{key} and \code{value} must be character strings. If \code{value} is missing, then the \code{header} value is returned.}
    \item{\code{redirect(target, status=302)}:}{ Sets up an HTTP redirect to the \code{target} url. }
    \item{\code{write(str)}:}{ Takes a character vector and appends it to the response body. }
    \item{\code{new(body='', status=200, headers=list())}:}{ Create a new Response object. \code{body} is a character vector, \code{status} is an HTTP status value. \code{headers} is a named list.}
    \item{\code{set_cookie(key, value)}:}{ Sets an HTTP cookie for the response. Both \code{key} and \code{value} must be character strings.}
    \item{\code{delete_cookie(key, value)}:}{ Sends appropriate HTTP header to delete the associated cookie on the client. \code{key} and \code{value} must be character strings.}
    \item{\code{finish()}:}{ Returns the response according to the Rook specification.}
  }
}

\name{Utils-class}
\Rdversion{1.1}
\docType{class}
\alias{Utils-class}
\alias{Utils}

\title{Class \code{Utils}}
\description{
A convenience object for working with various aspects of web requests and responses.
}
\seealso{
\code{\link{Multipart}}.
}
\examples{
Utils$bytesize('foo')
Utils$escape('foo bar')
Utils$unescape('foo+bar')
Utils$escape_html('foo <bar>')
Utils$escape('foo <bar>')
Utils$escape('foo\n<bar>')
Utils$status_code('OK')
Utils$status_code('Found')
Utils$status_code('Not Found')
x <- Utils$parse_query('foo=1&bar=baz')
x
Utils$rfc2822(Sys.time())
Utils$timezero()
Utils$build_query(x)
rm(x)
}
\keyword{classes}
\section{Methods}{
  \describe{
    \item{\code{bytesize(string=NULL)}:}{ Returns size in bytes for \code{string}, a character vector. }
    \item{\code{unescape(s=NULL)}:}{ returns the url decoded value of the character vector \code{s}. Also replaces the \code{'+'} character with a space. }
    \item{\code{status_code(status=NULL)}:}{ returns integer value for the given HTTP \code{status}, which can either be numeric or or a character vector describing the status. Returns \code{as.integer(500)} if \code{status} is NULL.}
    \item{\code{escape_html(string=NULL)}:}{ replaces \code{"&"}, \code{"<"}, \code{">"}, \code{"'"}, and \code{'"'} with entity equivalents. }
    \item{\code{raw.match(needle=NULL, haystack=NULL, all=TRUE)}:}{ returns index position of \code{needle} in  \code{haystack}. All matched indexes are returned by default. \code{needle} is either a raw vector or character string. \code{haystack} is a raw vector.}
    \item{\code{parse_query(qs=NULL, d=DEFAULT_SEP)}:}{ Creates a named list from the the query string \code{qs}. \code{d} is the separator value and defaults to \code{'[&;] *'}.}
    \item{\code{rfc2822(ts=NULL)}:}{ Formats \code{ts} in RFC2822 time. \code{ts} must be a \code{\link{POSIXt}} object.}
    \item{\code{escape(s=NULL)}:}{ Transforms any non-printable characters found in \code{s} to their percent-encoded equivalents.}
    \item{\code{build_query(params=NULL)}:}{ Creates a query string from the named list given in \code{params}. }
    \item{\code{timezero()}:}{ Returns a \code{POSIXct} object set to UNIX epoch. }
    \item{\code{set_cookie_header(header, key, value, expires, path, domain, secure, httpOnly)}:}{ Sets an HTTP cookie header in the environment \code{header}. All arguments except \code{expires} are length 1 character vectors, while \code{expires} must be a \code{POSIXct} object. }
    \item{\code{delete_cookie_header(header, key, value, expires, path, domain, secure, httpOnly)}:}{ Deletes the HTTP cookie header. }
  }
}

\name{Rook-package}
\alias{Rook-package}
\alias{Rook}
\docType{class}
\title{
Rook: A web server interface and package for R
}
\description{
    This help page defines the Rook specification. It borrows heavily
    from Ruby's Rack project: \url{https://github.com/rack/rack}.

    After reading this document, read the \code{\link{Rhttpd}} help file
    as it will get you familiar with installing and running \code{Rook}
    applications. Then explore the example applications located in:

    \code{system.file('exampleApps',package='Rook')}.
}

\section{Rook applications}{
A Rook application is an R reference class object that implements a 'call'
method or an R closure that takes exactly one argument, an environment,
and returns a list with three named elements: \code{'status'}, \code{'headers'},
and \code{'body'}.
}
\section{Hello World}{

Here is a basic Rook application as a closure that implements 'hello world':

\preformatted{
function(env){
    body = paste('<h1>Hello World! This is Rook',env$rook.version,'.</h1>')
    list(
        status = 200L,
        headers = list(
            'Content-Type' = 'text/html'
        ),
        body = body
    )
}
}

And the equivalent reference class example:

\preformatted{
setRefClass(
  'HelloWorld',
  methods = list(
    call = function(env){
      list(
        status = 200L,
        headers = list(
         'Content-Type' = 'text/html'
        ),
        body = paste('<h1>Hello World! This is Rook',env$rook.version,'.</h1>')
      )
    }
  )
)
}
}

\section{The Environment}{

The environment argument is a true R environment object which the
application is free to modify. It is required to contain the following
variables:
\describe{
  \item{REQUEST_METHOD}{ The HTTP request method, such as "GET" or "POST". This
    cannot ever be an empty string, and so is always required. }

  \item{SCRIPT_NAME}{ The initial portion of the request URL's "path" that
    corresponds to the application object, so that the application knows
    its virtual "location". This may be an empty string, if the application
    corresponds to the "root" of the server.}

  \item{PATH_INFO}{ The remainder of the request URL's "path", designating the
    virtual "location" of the request's target within the application. This
    may be an empty string, if the request URL targets the application root
    and does not have a trailing slash. This value may be percent-encoded
    when I originating from a URL.}

  \item{QUERY_STRING}{ The portion of the request URL that follows the ?,
    if any. May be empty, but is always required!}

\item{SERVER_NAME, SERVER_PORT}{ When combined with SCRIPT_NAME and PATH_INFO,
    these variables can be used to complete the URL. Note however that
    HTTP_HOST, if present, should be used in preference to SERVER_NAME for
    reconstructing the request URL. SERVER_NAME and SERVER_PORT can never
    be empty strings, and so are always required.}

\item{HTTP_ Variables}{ Variables corresponding to the client-supplied
    HTTP request headers (i.e., variables whose names begin with HTTP_). The
    presence or absence of these variables should correspond with the presence
    or absence of the appropriate HTTP header in the request.}
}

In addition, the environment must include the following Rook-specific variables:

\describe{
 \item{rook.version}{ This version of Rook.}
 \item{rook.url_scheme}{'http' or 'https', depending on the request URL.}
 \item{rook.input}{See \dQuote{The Input Stream} section.}
 \item{rook.errors}{See \dQuote{The Error Stream} section.}
}
}
\section{The Input Stream}{
The rook.input variable must contain an object created from a reference
class that implements \code{read_lines()}, \code{read()}, and \code{rewind()}:

\describe{
    \item{\code{read_lines(l=-1L)}:}{takes one argument, the number of lines to read. Includes partial ending line.}
    \item{\code{read(l=-1L)}:}{takes one argument, the number of bytes to read. Returns a raw vector.}
    \item{\code{rewind()}:}{Rewinds the input stream back to the beginning.}
}
}

\section{The Error Stream}{
The rook.error variable must contain an object created from a reference
class that implements \code{flush()} and \code{cat()}:
\describe{
    \item{\code{flush()}:}{called with no arguments and makes the error stream immediately appear.}
    \item{\code{cat(...,sep=" ",fill=FALSE,labels=NULL)}:}{called with the same arguments as R's \code{"\link[base]{cat}"} without the \code{file} and append \code{argument}.}
}
}

\section{The Response}{

Rook applications return a list with three named elements: \code{'status'}, \code{'headers'},
and \code{'body'}.

\subsection{\code{'status'}}{
An HTTP status value as integer and must be greater than or equal to 100. 
}

\subsection{\code{'headers'}}{
A named list that contains only character values corresponding to valid HTTP headers.
}
\subsection{\code{'body'}}{
Either a character or raw vector. If the character vector is named
with value \code{'file'} then value of the vector is interpreted as the
location of a file.
}}
\author{ Jeffrey Horner <jeffrey.horner@gmail.com> }
\keyword{package}

\name{RhttpdInputStream-class}
\Rdversion{1.1}
\docType{class}
\alias{RhttpdInputStream-class}
\alias{RhttpdInputStream}

\title{Class \code{RhttpdInputStream}}
\description{
An internal class used by \code{\link{Rhttpd}}.
}
\examples{
showClass("RhttpdInputStream")
}
\keyword{classes}

\name{is_rookable}
\alias{is_rookable}
\title{
Test for Rookable applications
}
\description{
A convenience function for testing whether or not objects are either a function or reference class as defined by the Rook specification for applications.
}
\usage{
is_rookable(app)
}
\arguments{
  \item{app}{ Any R object.}
}
\value{
    Logical determining whether or not argument is Rookable. Not vectorized.
}
\seealso{
\code{\link{Rook}}.
}
\keyword{function}

\name{Redirect-class}
\Rdversion{1.1}
\docType{class}
\alias{Redirect-class}
\alias{Redirect}

\title{Class \code{Redirect}}
\description{
A \code{\link{Rook}} application whose only role is to return an HTTP redirect header to the given url.
}
\seealso{
See \code{\link{Brewery}} for an example. 
}
\keyword{classes}
\section{Methods}{
  \describe{
    \item{\code{new(url)}:}{ Returns a \code{Rook} object. \code{url} is a character string whose value is a full or relative url to which the browser is redirected.}
  }
}

\name{App-class}
\Rdversion{1.1}
\docType{class}
\alias{App-class}
\alias{App}

\title{Class \code{App}}
\description{
Abstract class from which \code{Middleware} and \code{Builder} inherit. Provides the \code{app} field.

\code{App} can also be used to instantiate reference classed applications wrapped around a function. See \code{\link{Middleware}} for an example.
}
\section{Fields}{
  \describe{
    \item{\code{app}:}{A Rook application.}
  }
}
\section{Methods}{
  \describe{
    \item{\code{new(app=NULL)}:}{ Creates a new \code{App} object. \code{app} is any \code{Rook} aware R object.}
  }
}
\seealso{
\code{\link{is_rookable}}, \code{\link{Builder}}, and \code{\link{Middleware}}.
}
\keyword{classes}

\name{Static-class}
\Rdversion{1.1}
\docType{class}
\alias{Static-class}
\alias{Static}

\title{Class \code{Static}}
\description{
A \code{\link{Middleware}} class for serving static files from a root directory given a set of url paths.
}
\seealso{
See \code{\link{Builder}} for an example.
}
\keyword{classes}
\section{Methods}{
  \describe{
    \item{\code{new(urls, root)}:}{ Creates a new object. \code{urls} is a character vector whose elements must start with a \code{'/'}. \code{root} is a length 1 character vector whose value must be a valid directory. }
  }
}

\name{Brewery-class}
\Rdversion{1.1}
\docType{class}
\alias{Brewery-class}
\alias{Brewery}

\title{Class \code{Brewery}}
\description{
A \code{\link{Middleware}} class for mapping URLs to a directory of files that are subsequently passed to \code{\link[brew]{brew}}. When a file is brewed, the two variables \code{req} (an object of class \code{\link{Request}}) and \code{res} (an object of class \code{\link{Response}}) are available for use.

}
\section{Methods}{
\describe{
    \item{\code{new(url,root,...):}}{
	\code{url} is a character string or \code{\link{regexp}} on which to match, \code{root} is the name of the directory where brew files reside. Named arguments can be passed in via \code{...} and will be available within the scope of each brewed file.
    }
}}

\examples{

#
# This application runs any file found in tempdir() through brew.
#
s <- Rhttpd$new()
\dontrun{
s$start(quiet=TRUE)
}
cat("<h1>Random Number: <\%=rnorm(1)\%></h1>",
    file=file.path(tempdir(),"index.html"))
s$add(name="random",
    app=Builder$new(
        Brewery$new(url="/",root=tempdir()),
        Redirect$new("/index.html")
    )
)
\dontrun{
s$browse('random') # Opens a browser window to the app.
}
file.remove(file.path(tempdir(),"index.html"))
s$remove(all=TRUE)
rm(s)
}

\seealso{
\code{\link{Rhttpd}}, \code{\link{Builder}}, \code{\link{Redirect}}, and \code{\link[brew]{brew}}.
}
\keyword{classes}

\name{URLMap-class}
\Rdversion{1.1}
\docType{class}
\alias{URLMap-class}
\alias{URLMap}

\title{Class \code{URLMap}}
\description{
A \code{\link{Rook}} application that maps url paths to other \code{Rook} applications.
}
\seealso{ 
\code{\link{Rhttpd}}.
}
\examples{
s <- Rhttpd$new()
s$add(
    name="pingpong",
    app=Rook::URLMap$new(
        '/ping' = function(env){
            req <- Rook::Request$new(env)
            res <- Rook::Response$new()
            res$write(sprintf('<h1><a href="\%s">Pong</a></h1>',req$to_url("/pong")))
            res$finish()
        },
        '/pong' = function(env){
            req <- Rook::Request$new(env)
            res <- Rook::Response$new()
            res$write(sprintf('<h1><a href="\%s">Ping</a></h1>',req$to_url("/ping")))
            res$finish()
        },
        '/?' = function(env){
            req <- Rook::Request$new(env)
            res <- Rook::Response$new()
            res$redirect(req$to_url('/pong'))
            res$finish()
        }
    )
)
\dontrun{
s$start(quiet=TRUE)
s$browse('pingpong')
}
s$remove('pingpong')
\dontrun{
s$stop()
}
rm(s)
}
\keyword{classes}
\section{Methods}{
  \describe{
    \item{\code{new(...)}:}{ Creates a \code{Rook} application. All arguments must be \code{Rook} applications and named as in the example.}
  }
}

\name{Builder-class}
\Rdversion{1.1}
\docType{class}
\alias{Builder-class}
\alias{Builder}

\title{Class \code{Builder}}
\description{
A convenience object for combining various \code{Middleware} with a default application to create a more complex Rook application.
}
\examples{

# The following is the Hmisc example. Explore the folder
# system.file('exampleApps/Hmisc',package='Rook') for more information.
s <- Rhttpd$new()
\dontrun{
library(Hmisc)
dir.create(file.path(tempdir(),'plots'),showWarnings=FALSE)
s$add( name="Hmisc",
    app=Builder$new(
        Static$new(
            urls = c('/css','/images','/javascript'),
            root = system.file('exampleApps/Hmisc',package='Rook')
        ),
        Static$new(urls='/plots',root=tempdir()),
        Brewery$new(
            url='/brew',
            root= system.file('exampleApps/Hmisc',package='Rook'),
            imagepath=file.path(tempdir(),'plots'),
            imageurl='../plots/'
        ),
        Redirect$new('/brew/useR2007.rhtml')
    )
)
s$start(quiet=TRUE)
s$browse('Hmisc') # Opens a browser window to the application.
s$remove(all=TRUE)
s$stop()
}
}
\seealso{
\code{\link{Rhttpd}}, \code{\link{Static}}, \code{\link{Brewery}}, and \code{\link{Redirect}}.
}
\keyword{classes}
\section{Methods}{
  \describe{
    \item{\code{new(...)}:}{ Arguments can be any \code{Middleware} object while the last argument in the list must be a valid \code{Rook} application. That is, it will handle the incoming request without deferring to another application. }
  }
}
